"""
Functions for converting dates to/from JD and MJD. Assumes dates are historical
dates, including the transition from the Julian calendar to the Gregorian
calendar in 1582. No support for proleptic Gregorian/Julian calendars.

:Author: Matt Davis
:Website: http://github.com/jiffyclub

"""

import math
import datetime as dt

# Note: The Python datetime module assumes an infinitely valid Gregorian calendar.
#       The Gregorian calendar took effect after 10-15-1582 and the dates 10-05 through
#       10-14-1582 never occurred. Python datetime objects will produce incorrect
#       time deltas if one date is from before 10-15-1582.

def mjd_to_jd(mjd):
    """
    Convert Modified Julian Day to Julian Day.
        
    Parameters
    ----------
    mjd : float
        Modified Julian Day
        
    Returns
    -------
    jd : float
        Julian Day
    
        
    """
    return mjd + 2400000.5

    
def jd_to_mjd(jd):
    """
    Convert Julian Day to Modified Julian Day
    
    Parameters
    ----------
    jd : float
        Julian Day
        
    Returns
    -------
    mjd : float
        Modified Julian Day
    
    """
    return jd - 2400000.5

    
def date_to_jd(year,month,day):
    """
    Convert a date to Julian Day.
    
    Algorithm from 'Practical Astronomy with your Calculator or Spreadsheet', 
        4th ed., Duffet-Smith and Zwart, 2011.
    
    Parameters
    ----------
    year : int
        Year as integer. Years preceding 1 A.D. should be 0 or negative.
        The year before 1 A.D. is 0, 10 B.C. is year -9.
        
    month : int
        Month as integer, Jan = 1, Feb. = 2, etc.
    
    day : float
        Day, may contain fractional part.
    
    Returns
    -------
    jd : float
        Julian Day
        
    Examples
    --------
    Convert 6 a.m., February 17, 1985 to Julian Day
    
    >>> date_to_jd(1985,2,17.25)
    2446113.75
    
    """
    if month == 1 or month == 2:
        yearp = year - 1
        monthp = month + 12
    else:
        yearp = year
        monthp = month
    
    # this checks where we are in relation to October 15, 1582, the beginning
    # of the Gregorian calendar.
    if ((year < 1582) or
        (year == 1582 and month < 10) or
        (year == 1582 and month == 10 and day < 15)):
        # before start of Gregorian calendar
        B = 0
    else:
        # after start of Gregorian calendar
        A = math.trunc(yearp / 100.)
        B = 2 - A + math.trunc(A / 4.)
        
    if yearp < 0:
        C = math.trunc((365.25 * yearp) - 0.75)
    else:
        C = math.trunc(365.25 * yearp)
        
    D = math.trunc(30.6001 * (monthp + 1))
    
    jd = B + C + D + day + 1720994.5
    
    return jd
    
    
def jd_to_date(jd):
    """
    Convert Julian Day to date.
    
    Algorithm from 'Practical Astronomy with your Calculator or Spreadsheet', 
        4th ed., Duffet-Smith and Zwart, 2011.
    
    Parameters
    ----------
    jd : float
        Julian Day
        
    Returns
    -------
    year : int
        Year as integer. Years preceding 1 A.D. should be 0 or negative.
        The year before 1 A.D. is 0, 10 B.C. is year -9.
        
    month : int
        Month as integer, Jan = 1, Feb. = 2, etc.
    
    day : float
        Day, may contain fractional part.
        
    Examples
    --------
    Convert Julian Day 2446113.75 to year, month, and day.
    
    >>> jd_to_date(2446113.75)
    (1985, 2, 17.25)
    
    """
    jd = jd + 0.5
    
    F, I = math.modf(jd)
    I = int(I)
    
    A = math.trunc((I - 1867216.25)/36524.25)
    
    if I > 2299160:
        B = I + 1 + A - math.trunc(A / 4.)
    else:
        B = I
        
    C = B + 1524
    
    D = math.trunc((C - 122.1) / 365.25)
    
    E = math.trunc(365.25 * D)
    
    G = math.trunc((C - E) / 30.6001)
    
    day = C - E + F - math.trunc(30.6001 * G)
    
    if G < 13.5:
        month = G - 1
    else:
        month = G - 13
        
    if month > 2.5:
        year = D - 4716
    else:
        year = D - 4715
        
    return year, month, day
    
    
def hmsm_to_days(hour=0,min=0,sec=0,micro=0):
    """
    Convert hours, minutes, seconds, and microseconds to fractional days.
    
    Parameters
    ----------
    hour : int, optional
        Hour number. Defaults to 0.
    
    min : int, optional
        Minute number. Defaults to 0.
    
    sec : int, optional
        Second number. Defaults to 0.
    
    micro : int, optional
        Microsecond number. Defaults to 0.
        
    Returns
    -------
    days : float
        Fractional days.
        
    Examples
    --------
    >>> hmsm_to_days(hour=6)
    0.25
    
    """
    days = sec + (micro / 1.e6)
    
    days = min + (days / 60.)
    
    days = hour + (days / 60.)
    
    return days / 24.
    
    
def days_to_hmsm(days):
    """
    Convert fractional days to hours, minutes, seconds, and microseconds.
    Precision beyond microseconds is rounded to the nearest microsecond.
    
    Parameters
    ----------
    days : float
        A fractional number of days. Must be less than 1.
        
    Returns
    -------
    hour : int
        Hour number.
    
    min : int
        Minute number.
    
    sec : int
        Second number.
    
    micro : int
        Microsecond number.
        
    Raises
    ------
    ValueError
        If `days` is >= 1.
        
    Examples
    --------
    >>> days_to_hmsm(0.1)
    (2, 24, 0, 0)
    
    """
    hours = days * 24.
    hours, hour = math.modf(hours)
    
    mins = hours * 60.
    mins, min = math.modf(mins)
    
    secs = mins * 60.
    secs, sec = math.modf(secs)
    
    micro = round(secs * 1.e6)
    
    return int(hour), int(min), int(sec), int(micro)
    

def datetime_to_jd(date):
    """
    Convert a `datetime.datetime` object to Julian Day.
    
    Parameters
    ----------
    date : `datetime.datetime` instance
    
    Returns
    -------
    jd : float
        Julian day.
        
    Examples
    --------
    >>> d = datetime.datetime(1985,2,17,6)  
    >>> d
    datetime.datetime(1985, 2, 17, 6, 0)
    >>> jdutil.datetime_to_jd(d)
    2446113.75
    
    """
    days = date.day + hmsm_to_days(date.hour,date.minute,date.second,date.microsecond)
    
    return date_to_jd(date.year,date.month,days)
    
    
def jd_to_datetime(jd):
    """
    Convert a Julian Day to an `jdutil.datetime` object.
    
    Parameters
    ----------
    jd : float
        Julian day.
        
    Returns
    -------
    dt : `jdutil.datetime` object
        `jdutil.datetime` equivalent of Julian day.
    
    Examples
    --------
    >>> jd_to_datetime(2446113.75)
    datetime(1985, 2, 17, 6, 0)
    
    """
    year, month, day = jd_to_date(jd)
    
    frac_days,day = math.modf(day)
    day = int(day)
    
    hour,min,sec,micro = days_to_hmsm(frac_days)
    
    return datetime(year,month,day,hour,min,sec,micro)


def timedelta_to_days(td):
    """
    Convert a `datetime.timedelta` object to a total number of days.
    
    Parameters
    ----------
    td : `datetime.timedelta` instance
    
    Returns
    -------
    days : float
        Total number of days in the `datetime.timedelta` object.
        
    Examples
    --------
    >>> td = datetime.timedelta(4.5)
    >>> td
    datetime.timedelta(4, 43200)
    >>> timedelta_to_days(td)
    4.5
    
    """
    seconds_in_day = 24. * 3600.
    
    days = td.days + (td.seconds + (td.microseconds * 10.e6)) / seconds_in_day
    
    return days
    
    
class datetime(dt.datetime):
    """
    A subclass of `datetime.datetime` that performs math operations by first
    converting to Julian Day, then back to a `jdutil.datetime` object.
    
    Addition works with `datetime.timedelta` objects, subtraction works with
    `datetime.timedelta`, `datetime.datetime`, and `jdutil.datetime` objects.
    Not all combinations work in all directions, e.g.
    `timedelta - datetime` is meaningless.
    
    See Also
    --------
    datetime.datetime : Parent class.
    
    """
    def __add__(self,other):
        if not isinstance(other,dt.timedelta):
            s = "jdutil.datetime supports '+' only with datetime.timedelta"
            raise TypeError(s)
        
        days = timedelta_to_days(other)
        
        combined = datetime_to_jd(self) + days
        
        return jd_to_datetime(combined)
        
    def __radd__(self,other):
        if not isinstance(other,dt.timedelta):
            s = "jdutil.datetime supports '+' only with datetime.timedelta"
            raise TypeError(s)
        
        days = timedelta_to_days(other)
        
        combined = datetime_to_jd(self) + days
        
        return jd_to_datetime(combined)
        
    def __sub__(self,other):
        if isinstance(other,dt.timedelta):
            days = timedelta_to_days(other)
            
            combined = datetime_to_jd(self) - days
            
            return jd_to_datetime(combined)
            
        elif isinstance(other, (datetime,dt.datetime)):
            diff = datetime_to_jd(self) - datetime_to_jd(other)
            
            return dt.timedelta(diff)
            
        else:
            s = "jdutil.datetime supports '-' with: "
            s += "datetime.timedelta, jdutil.datetime and datetime.datetime"
            raise TypeError(s)
            
    def __rsub__(self,other):
        if not isinstance(other, (datetime,dt.datetime)):
            s = "jdutil.datetime supports '-' with: "
            s += "jdutil.datetime and datetime.datetime"
            raise TypeError(s)
            
        diff = datetime_to_jd(other) - datetime_to_jd(self)
            
        return dt.timedelta(diff)
        
    def to_jd(self):
        """
        Return the date converted to Julian Day.
        
        """
        return datetime_to_jd(self)
        
    def to_mjd(self):
        """
        Return the date converted to Modified Julian Day.
        
        """
        return jd_to_mjd(self.to_jd())
    


import numpy as np
import glob
import os

# import json
# import pickle
# from functools import wraps
# import astropy.units as u

__author__ = ["Sijie Yu"]
__email__ = "sijie.yu@njit.edu"


def insertchar(source_str, insert_str, pos):
    return source_str[:pos] + insert_str + source_str[pos:]


def get_curve_grad(x, y):
    '''
    get the grad of at data point
    :param x:
    :param y:
    :return: grad,posang
    '''
    deltay = np.roll(y, -1) - np.roll(y, 1)
    deltay[0] = y[1] - y[0]
    deltay[-1] = y[-1] - y[-2]
    deltax = np.roll(x, -1) - np.roll(x, 1)
    deltax[0] = x[1] - x[0]
    deltax[-1] = x[-1] - x[-2]
    grad = deltay / deltax
    posang = np.arctan2(deltay, deltax)
    return {'grad': grad, 'posang': posang}


def findDist(x, y):
    dx = np.diff(x)
    dy = np.diff(y)
    dist = np.hypot(dx, dy)
    return np.insert(dist, 0, 0.0)


def paramspline(x, y, length, s=0):
    from scipy.interpolate import splev, splprep
    tck, u = splprep([x, y], s=s)
    unew = np.linspace(0, u[-1], length)
    out = splev(unew, tck)
    xs, ys = out[0], out[1]
    grads = get_curve_grad(xs, ys)
    return {'xs': xs, 'ys': ys, 'grads': grads['grad'], 'posangs': grads['posang']}


def polyfit(x, y, length, deg, keepxorder=False):
    T = False
    if len(x) == 2:
        if np.abs(np.diff(y) / np.diff(x)) < 1:
            xproc = x
            yproc = y
        else:
            xproc = y
            yproc = x
            T = True
    else:
        xproc = x
        yproc = y
    if keepxorder:
        xs = np.linspace(xproc[0], xproc[-1], length)
    else:
        xs = np.linspace(np.nanmin(xproc), np.nanmax(xproc), length)
    z = np.polyfit(x=xproc, y=yproc, deg=deg)
    p = np.poly1d(z)
    ys = p(xs)
    grads = get_curve_grad(xs, ys)
    if T:
        return {'xs': ys, 'ys': xs, 'grads': 1/grads['grad'], 'posangs': np.pi/2-grads['posang']}
    else:
        return {'xs': xs, 'ys': ys, 'grads': grads['grad'], 'posangs': grads['posang']}


def improfile(z, xi, yi, interp='cubic'):
    '''
    Pixel-value cross-section along line segment in an image
    :param z: an image array
    :param xi and yi: equal-length vectors specifying the pixel coordinates of the endpoints of the line segment
    :param interp: interpolation type to sampling, 'nearest' or 'cubic'
    :return: the intensity values of pixels along the line
    '''
    import scipy.ndimage
    imgshape = z.shape
    if len(xi) != len(yi):
        raise ValueError('xi and yi must be equal-length!')
    if len(xi) < 2:
        raise ValueError('xi or yi must contain at least two elements!')
    for idx, ll in enumerate(xi):
        if not 0 < ll < imgshape[1]:
            return np.nan
        if not 0 < yi[idx] < imgshape[0]:
            return np.nan
    if len(xi) == 2:
        length = np.floor(np.hypot(np.diff(xi), np.diff(yi))[0]).astype(np.int)
        x, y = np.linspace(xi[0], xi[1], length), np.linspace(yi[0], yi[1], length)
    else:
        x, y = xi, yi
    if interp == 'cubic':
        zi = scipy.ndimage.map_coordinates(z, np.vstack((x, y)))
    else:
        zi = z[np.floor(y).astype(np.int), np.floor(x).astype(np.int)]

    return zi


def map2wcsgrids(snpmap, cell=True, antialiased=True):
    '''

    :param snpmap:
    :param cell: if True, return the coordinates of the pixel centers. if False, return the coordinates of the pixel boundaries
    :return:
    '''
    # embed()
    import astropy.units as u
    ny, nx = snpmap.data.shape
    x0, x1 = snpmap.xrange.to(u.arcsec).value
    y0, y1 = snpmap.yrange.to(u.arcsec).value
    dx = snpmap.scale.axis1.to(u.arcsec / u.pix).value
    dy = snpmap.scale.axis2.to(u.arcsec / u.pix).value

    if cell:
        mapx, mapy = np.linspace(x0, x1, nx), np.linspace(y0, y1, ny)
        mapx = np.tile(mapx, ny).reshape(ny, nx)
        mapy = np.tile(mapy, nx).reshape(nx, ny).transpose()
    else:
        nx += 1
        ny += 1
        mapx, mapy = np.linspace(x0 - dx / 2.0, x1 + dx / 2.0, nx), np.linspace(y0 - dy / 2.0, y1 + dy / 2.0, ny)
        mapx = np.tile(mapx, ny).reshape(ny, nx)
        mapy = np.tile(mapy, nx).reshape(nx, ny).transpose()
    return mapx, mapy


def readsdofile(datadir=None, wavelength=None, trange=None, isexists=False, timtol=1):
    '''
    read sdo file from local database
    :param datadir:
    :param wavelength:
    :param trange: the timestamp or timerange in Julian days. if is timerange, return a list of files in the timerange
    :param isexists: check if file exist. if files exist, return file name
    :param timtol: time difference tolerance in days for considering data as the same timestamp
    :return:
    '''
    from astropy.time import Time
    import sunpy.map
    from datetime import date
    from datetime import timedelta as td

    trange = Time(trange)
    wavelength = str(wavelength)
    wavelength = wavelength.lower()
    if timtol < 12. / 3600 / 24:
        timtol = 12. / 3600 / 24

    if len(trange.iso) == 2:
        if trange[1] < trange[0]:
            raise ValueError('start time must be occur earlier than end time!')
        else:
            sdofitspath = []
            jdtimestr = [Time(ll, format='jd').iso for ll in trange]
            ymd = [ll.split(' ')[0].split('-') for ll in jdtimestr]
            d1 = date(int(ymd[0][0]), int(ymd[0][1]), int(ymd[0][2]))
            d2 = date(int(ymd[1][0]), int(ymd[1][1]), int(ymd[1][2]))
            delta = d2 - d1
            for i in range(delta.days + 1):
                ymd = d1 + td(days=i)
                sdofitspathtmp = glob.glob(
                    datadir + '/{:04d}/{:02d}/{:02d}/aia.lev1_*Z.{}.image_lev1.fits'.format(ymd.year, ymd.month,
                                                                                            ymd.day, wavelength))
                if len(sdofitspathtmp) > 0:
                    sdofitspath = sdofitspath + sdofitspathtmp
            if len(sdofitspath) == 0:
                if isexists:
                    return sdofitspath
                else:
                    raise ValueError(
                        'No SDO file found under {} at the time range of {} to {}. Download the data with EvtBrowser first.'.format(
                            datadir,
                            jdtimestr[0],
                            jdtimestr[1]))
            sdofits = [os.path.basename(ll) for ll in sdofitspath]
            sdotimeline = Time(
                [insertchar(insertchar(ll.split('.')[2].replace('T', ' ').replace('Z', ''), ':', -4), ':', -2) for ll in
                 sdofits],
                format='iso', scale='utc')
            sdofitspathnew = [x for (y, x) in sorted(zip(sdotimeline.jd, sdofitspath))]
            sdofitsnew = [os.path.basename(ll) for ll in sdofitspathnew]
            sdotimelinenew = Time(
                [insertchar(insertchar(ll.split('.')[2].replace('T', ' ').replace('Z', ''), ':', -4), ':', -2) for ll in
                 sdofitsnew], format='iso',
                scale='utc')
            sdofile = list(
                np.array(sdofitspathnew)[
                    np.where(np.logical_and(trange[0].jd <= sdotimelinenew.jd, sdotimelinenew.jd <= trange[1].jd))[0]])
            return sdofile
    else:
        jdtimstr = trange.iso
        ymd = jdtimstr.split(' ')[0].split('-')
        sdofitspath = glob.glob(
            datadir + '/{}/{}/{}/aia.lev1_*Z.{}.image_lev1.fits'.format(ymd[0], ymd[1], ymd[2], wavelength))
        if len(sdofitspath) == 0:
            return []  # raise ValueError('No SDO file found under {}.'.format(datadir))
        sdofits = [os.path.basename(ll) for ll in sdofitspath]
        sdotimeline = Time(
            [insertchar(insertchar(ll.split('.')[2].replace('T', ' ').replace('Z', ''), ':', -4), ':', -2) for ll in
             sdofits],
            format='iso', scale='utc')
        if timtol < np.min(np.abs(sdotimeline.jd - trange.jd)):
            return []  # raise ValueError('No SDO file found at the select timestamp. Download the data with EvtBrowser first.')
        idxaia = np.argmin(np.abs(sdotimeline.jd - trange.jd))
        sdofile = sdofitspath[idxaia]
        if isexists:
            return sdofile
        else:
            try:
                sdomap = sunpy.map.Map(sdofile)
                return sdomap
            except:
                raise ValueError('File not found or invalid input')


def get_map_corner_coord(sunpymap):
    bottom_left_coord = sunpymap.bottom_left_coord
    top_right_coord = sunpymap.top_right_coord
    x0, y0 = bottom_left_coord.Tx, bottom_left_coord.Ty
    x1, y1 = top_right_coord.Tx, top_right_coord.Ty
    unit = x0.unit
    return np.array([x0.value, x1.value, y0.value, y1.value]) * unit



import astropy.units as u
import matplotlib.cm as cm
import matplotlib.colors as colors
import numpy as np
import sunpy.coordinates
import sunpy.map
import sunpy.cm
import sunpy.wcs as wcs
from bokeh.models import ColumnDataSource
from bokeh.plotting import figure
from bokeh.models.mappers import LogColorMapper, LinearColorMapper
from skimage.io import imread
from suncasa.utils import DButil

__author__ = ["Sijie Yu"]
__email__ = "sijie.yu@njit.edu"

# define the colormaps
colormap_jet = cm.get_cmap("jet")  # choose any matplotlib colormap here
bokehpalette_jet = [colors.rgb2hex(m) for m in colormap_jet(np.arange(colormap_jet.N))]


def getAIApalette(wavelength):
    clmap = cm.get_cmap("sdoaia" + wavelength)  # choose any matplotlib colormap here
    palette = [colors.rgb2hex(m).encode("ascii").upper() for m in clmap(np.arange(clmap.N))]
    return palette


class PuffinMap:
    """
    Produce A Generic Column data source from map for Bokeh plot

    Parameters
    ----------
    map : Sunpy map
    """
    __slots__ = ['smap', 'plot_height', 'plot_width', 'x', 'y', 'dw', 'dh', 'x_range', 'y_range']

    def __init__(self, data=None, header=None, smap=None, plot_height=None, plot_width=None, *args,
                 **kwargs):
        if not smap:
            smap = sunpy.map.Map((data, header))
        self.smap = smap
        if not plot_height:
            plot_height = 400.0
        if not plot_width:
            plot_width = 400.0
        self.plot_height = plot_height
        self.plot_width = plot_width

        """
        define the plot region based the plot_width and plot_height
        ------------------
        x (DataSpecProperty) - The x-coordinates to locate the image anchors.
        y (DataSpecProperty) - The y-coordinates to locate the image anchors.
        dw (UnitsSpecProperty) - The widths of the plot regions that the images will occupy.
        dh (UnitsSpecProperty) - The height of the plot region that the image will occupy.
        """
        try:
            dx, dy = self.smap.scale.x.value, self.smap.scale.y.value
        except:
            dx, dy = self.smap.scale[0].value, self.smap.scale[1].value

        self.dx, self.dy = dx, dy
        x0, x1 = self.smap.xrange.value[0] - dx / 2.0, self.smap.xrange.value[1] + dx / 2.0
        y0, y1 = self.smap.yrange.value[0] - dy / 2.0, self.smap.yrange.value[1] + dy / 2.0
        self.x, self.y = [x0], [y0]
        self.dw, self.dh = [x1 - x0], [y1 - y0]
        self.x_range = [x0, x1]
        # self.y_range = [y0, y0 + (y1 - y0) * float(self.plot_height) / float(self.plot_width)]
        self.y_range = [y0, y1]

    def meshgrid(self, rescale=1.0, *args, **kwargs):
        XX, YY = np.meshgrid(np.arange(self.smap.data.shape[1] * rescale), np.arange(self.smap.data.shape[0] * rescale))
        mesh =  self.smap.pixel_to_world(XX / rescale * u.pix, YY / rescale * u.pix)
        x,y = mesh.Tx,mesh.Tx
        return x, y

    def meshgridpix(self, rescale=1.0, *args, **kwargs):
        XX, YY = np.meshgrid(np.arange(self.smap.data.shape[1] * rescale) / rescale,
                             np.arange(self.smap.data.shape[0] * rescale) / rescale)
        x, y = XX * u.pix, YY * u.pix
        return x, y

    def ImageSource(self, *args, **kwargs):
        """maps the Sunpy map to Bokeh DataSource
        """
        # XX, YY = (np.arange(self.smap.data.shape[0]), np.zeros(self.smap.data.shape[0]))
        # x = self.smap.pixel_to_data(XX * u.pix, YY * u.pix)[0]
        # XX, YY = (np.zeros(self.smap.data.shape[1]), np.arange(self.smap.data.shape[1]))
        # y = self.smap.pixel_to_data(XX * u.pix, YY * u.pix)[1]
        data = self.smap.data.copy()
        data[~np.isnan(data)] = data[~np.isnan(data)] / self.smap.exposure_time.value
        # return {'data': [data], 'xx': [x], 'yy': [y]}
        return {'data': [data]}

    def DrawGridSource(self, grid_spacing=15 * u.deg, *args, **kwargs):
        """maps the Longitude and Latitude grids to Bokeh DataSource
        """
        # XX, YY = np.meshgrid(np.arange(self.smap.data.shape[0]), np.arange(self.smap.data.shape[1]))
        # x, y = self.smap.pixel_to_data(XX * u.pix, YY * u.pix)
        dsun = self.smap.dsun

        b0 = self.smap.heliographic_latitude.to(u.deg).value
        l0 = self.smap.heliographic_longitude.to(u.deg).value
        units = self.smap.spatial_units

        xs = []
        ys = []

        hg_longitude_deg = np.linspace(-180, 180, num=361) + l0
        hg_latitude_deg = np.arange(-90, 90, grid_spacing.to(u.deg).value)

        # draw the latitude lines
        for lat in hg_latitude_deg:
            x, y = wcs.convert_hg_hpc(hg_longitude_deg, lat * np.ones(361), b0_deg=b0, l0_deg=l0, dsun_meters=dsun,
                                      angle_units=units.axis1, occultation=True)
            valid = np.logical_and(np.isfinite(x), np.isfinite(y))
            x = x[valid]
            y = y[valid]
            xs.append(x.tolist())
            ys.append(y.tolist())

        hg_longitude_deg = np.arange(-180, 180, grid_spacing.to(u.deg).value) + l0
        hg_latitude_deg = np.linspace(-90, 90, num=181)

        # draw the longitude lines
        for lon in hg_longitude_deg:
            x, y = wcs.convert_hg_hpc(lon * np.ones(181), hg_latitude_deg, b0_deg=b0, l0_deg=l0, dsun_meters=dsun,
                                      angle_units=units.axis1, occultation=True)
            valid = np.logical_and(np.isfinite(x), np.isfinite(y))
            x = x[valid]
            y = y[valid]
            xs.append(x.tolist())
            ys.append(y.tolist())

        return ColumnDataSource(data={'xs': xs, 'ys': ys})

    def DrawLimbSource(self, *args, **kwargs):
        """maps the solar limb to Bokeh DataSource
        """
        radius = self.smap.rsun_obs.value

        phi = np.linspace(-180, 180, num=361) * np.pi / 180.
        x = np.cos(phi) * radius
        y = np.sin(phi) * radius

        return ColumnDataSource(data={'x': x, 'y': y})

    def PlotMap(self, DrawLimb=True, DrawGrid=True, grid_spacing=15 * u.deg, ignore_coord=False, title=None, tools=None,
                x_range=None, y_range=None,
                palette=None, imagetype='image', *args,
                **kwargs):
        """Plot the map using the bokeh.plotting interface
        """

        source_image = self.ImageSource()
        if not title:
            title = self.smap.name

        if not tools:
            tools = 'pan,wheel_zoom,save,reset'
        if not x_range:
            x_range = self.x_range
        if not y_range:
            y_range = self.y_range

        if not palette:
            if self.smap.observatory == 'SDO' and self.smap.instrument[0:3] == 'AIA':
                wavelngth = '{:.0f}'.format(self.smap.wavelength.value)
                # clmap = cm.get_cmap("sdoaia" + wavelngth)  # choose any matplotlib colormap here
                # palette = [colors.rgb2hex(m).encode("ascii").upper() for m in clmap(np.arange(clmap.N))]
                palette = getAIApalette(wavelngth)
                clrange = DButil.sdo_aia_scale_dict(wavelength=wavelngth, imagetype=imagetype)
                print clrange
                if clrange['log']:
                    colormapper = LogColorMapper(palette=palette, low=clrange['low'], high=clrange['high'])
                else:
                    colormapper = LinearColorMapper(palette=palette, low=clrange['low'], high=clrange['high'])
            else:
                palette = bokehpalette_jet
                colormapper = LinearColorMapper(palette=palette)
        else:
            colormapper = LinearColorMapper(palette=palette)

        if ignore_coord:
            x_range = [0 - 0.5, self.smap.data.shape[0] + 0.5]
            y_range = [0 - 0.5, self.smap.data.shape[1] + 0.5]
            p_xaxis_visible = False
            p_yaxis_visible = False
            p_xaxis_axislabel = ''
            p_yaxis_axislabel = ''
            x0 = 0 - 0.5
            y0 = 0 - 0.5
            dw = self.smap.data.shape[0]
            dh = self.smap.data.shape[1]
        else:
            p_xaxis_visible = True
            p_yaxis_visible = True
            p_xaxis_axislabel = 'X-position [arcsec]'
            p_yaxis_axislabel = 'Y-position [arcsec]'
            x0 = self.x
            y0 = self.y
            dw = self.dw
            dh = self.dh

        p_image = figure(tools=tools, x_range=x_range,
                         y_range=y_range, title=title, plot_height=self.plot_height,
                         plot_width=self.plot_width, *args, **kwargs)
        p_image.xaxis.axis_label = p_xaxis_axislabel
        p_image.yaxis.axis_label = p_yaxis_axislabel
        p_image.xaxis.visible = p_xaxis_visible
        p_image.yaxis.visible = p_yaxis_visible
        r_img = p_image.image(image=source_image['data'], x=x0, y=y0, dw=dw, dh=dh, color_mapper=colormapper)
        if DrawLimb:
            p_image.line(x='x', y='y', line_color='white', line_dash='solid', source=self.DrawLimbSource())
            if DrawGrid:
                p_image.multi_line(xs='xs', ys='ys', line_color='white', line_dash='dotted',
                                   source=self.DrawGridSource(grid_spacing=grid_spacing))
        return p_image, r_img


class PuffinImage:
    """
    Produce a A Generic Column data source from image for Bokeh plot

    Parameters
    ----------
    image : image data
    imfile: image file
    """
    __slots__ = ['plot_height', 'plot_width', 'img', 'x', 'y', 'dw', 'dh', 'x_range', 'y_range']

    def __init__(self, image=None, imfile=None, plot_height=None, plot_width=None, *args, **kwargs):
        if image is None and imfile is None:
            raise IOError("Missing input image or image file")
        else:
            if imfile is None:
                im = image
            else:
                im = imread(imfile)
            if not plot_height:
                plot_height = im.shape[0]
            if not plot_width:
                plot_width = im.shape[1]

            self.plot_height = plot_height
            self.plot_width = plot_width
            img = np.empty((plot_width, plot_height), dtype=np.uint32)
            view = img.view(dtype=np.uint8).reshape((plot_width, plot_height, 4))
            view[:, :, 0:3] = im[:, :, 0:3]
            self.img = np.flipud(img)

        """
        defind the plot region based the plot_width and plot_height
        ------------------
        x (DataSpecProperty) - The x-coordinates to locate the image anchors.
        y (DataSpecProperty) - The y-coordinates to locate the image anchors.
        dw (UnitsSpecProperty) - The widths of the plot regions that the images will occupy.
        dh (UnitsSpecProperty) - The height of the plot region that the image will occupy.
        """

        self.x, self.y = 0, 0
        self.dw, self.dh = plot_width, plot_height
        self.x_range = [0, plot_width]
        self.y_range = [0, plot_height]

    def ImageSource(self, *args, **kwargs):
        """maps the jpg to Bokeh DataSource
        """
        x = np.arange(self.img.shape[0])
        y = np.arange(self.img.shape[1])
        return ColumnDataSource(data={'data': [self.img], 'xx': [x], 'yy': [y]})

    def PlotImage(self, title=None, x_range=None, y_range=None, *args, **kwargs):
        """Plot the image using the bokeh.plotting interface
        """

        source_image = self.ImageSource()
        if not title:
            title = ''

        if not x_range:
            x_range = self.x_range
        if not y_range:
            y_range = self.y_range

        # plot the global vla image
        p_image = figure(tools='', x_range=x_range, y_range=y_range, title=title, toolbar_location=None,
                         plot_height=self.plot_height, plot_width=self.plot_width)
        p_image.xaxis.visible = False
        p_image.yaxis.visible = False
        r_img = p_image.image_rgba(image="data", x=self.x, y=self.y, dw=self.dw, dh=self.dh, source=source_image)

        return p_image, r_img



import os
import numpy as np
from taskinit import tb, casalog
from concat_cli import concat_cli as concat
from clearcal_cli import clearcal_cli as clearcal
from split_cli import split_cli as split


def concateovsa(msname, msfiles, visprefix='./', doclearcal=True, keep_orig_ms=False,
                cols2rm=["MODEL_DATA", "CORRECTED_DATA"]):
    concatvis = visprefix + msname
    msfiles_ = []
    for idx, ll in enumerate(msfiles):
        if str(ll).endswith('/'):
            msfiles[idx] = str(ll)[:-1]
    if doclearcal:
        print('Warning: Corrected column in the input ms file will be cleared!!!')
        for ll in msfiles:
            clearcal(vis=str(ll), addmodel=True)
    else:
        try:
            tmpdir = visprefix + '/tmp_ms/'
            if not os.path.exists(tmpdir):
                os.makedirs(tmpdir)
            for ll in msfiles:
                msfile_ = tmpdir + os.path.basename(str(ll))
                msfiles_.append(msfile_)
                split(vis=str(ll), outputvis=msfile_, datacolumn='corrected')
                clearcal(vis=msfile_, addmodel=True)
        except:
            print('Warning: Corrected column not found in the input ms file.')
            msfiles_ = msfiles
    if msfiles_:
        concat(vis=msfiles_, concatvis=concatvis, timesort=True)
    else:
        concat(vis=msfiles, concatvis=concatvis, timesort=True)
    # Change all observation ids to be the same (zero)
    tb.open(concatvis + '/OBSERVATION', nomodify=False)
    nobs = tb.nrows()
    tim0 = tb.getcell('TIME_RANGE', 0)[0]
    tim1 = tb.getcell('TIME_RANGE', nobs - 1)[1]
    tb.removerows([i + 1 for i in range(nobs - 1)])
    tb.putcell('TIME_RANGE', 0, [tim0, tim1])
    tb.close()

    tb.open(concatvis + '/DATA_DESCRIPTION', nomodify=False)
    nrows = tb.nrows()
    pol_id = tb.getcol('POLARIZATION_ID')
    tb.removerows(np.where(pol_id !=0)[0])
    tb.close()

    tb.open(concatvis, nomodify=False)
    dd_id = tb.getcol('DATA_DESC_ID')
    idx_dd_id, = np.where(dd_id >=nrows/2)
    dd_id[idx_dd_id] = dd_id[idx_dd_id] - nrows/2
    tb.putcol('DATA_DESC_ID',dd_id)
    tb.close()

    tb.open(concatvis + '/FIELD', nomodify=False)
    nobs = tb.nrows()
    tb.removerows([i + 1 for i in range(nobs - 1)])
    tb.close()

    tb.open(concatvis + '/SOURCE', nomodify=False)
    nobs = tb.nrows()
    tb.removerows([i + 1 for i in range(nobs - 1)])
    tb.close()

    tb.open(concatvis, nomodify=False)
    obsid = tb.getcol('OBSERVATION_ID')
    newobsid = np.zeros(len(obsid), dtype='int')
    tb.putcol('OBSERVATION_ID', newobsid)
    fldid = tb.getcol('FIELD_ID')
    newfldid = np.zeros(len(fldid), dtype='int')
    tb.putcol('FIELD_ID', newfldid)
    colnames = tb.colnames()
    for l in range(len(cols2rm)):
        if cols2rm[l] in colnames:
            try:
                tb.removecols(cols2rm[l])
            except:
                pass
    tb.close()



    if msfiles_ != [] and msfiles_ != msfiles:
        for ll in msfiles_:
            os.system('rm -rf {}'.format(ll))
    if not keep_orig_ms:
        for ll in msfiles:
            os.system('rm -rf {}'.format(ll))



import matplotlib
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.colorbar as colorbar
import matplotlib.cm as cm
import matplotlib.colors as colors
import astropy.units as u
from astropy.io import fits
from astropy.time import Time
import numpy as np
import sunpy.map
from suncasa.utils import DButil


def add_scalebar(length, x, y, ax, align='right', label='', label_position='below', yoff=-0.005, xoff=None, fontsize=None, **kwargs):
    '''
    :param length : length of scalebar in data coordinate
    :param x,y : in the coordinate system of the Axes; (0,0) is bottom left of the axes, and (1,1) is top right of the axes.
    :param ax:
    :param text:
    :param yoff : Yoffset of the label in the coordinate system of the Axes
    :return:
    '''
    x, y = ax.transData.inverted().transform(ax.transAxes.transform([x, y]))
    if align == 'right':
        x0, y0 = x - length, y
        x1, y1 = x, y
    elif align == 'left':
        x0, y0 = x, y
        x1, y1 = x + length, y
    elif align == 'middle':
        x0, y0 = x - length / 2.0, y
        x1, y1 = x + length / 2.0, y
    kwargs_text = {}
    if 'c' or 'color' in kwargs.keys():
        try:
            kwargs_text['color'] = kwargs['c']
        except:
            kwargs_text['color'] = kwargs['color']
    else:
        kwargs['c'] = 'white'
        kwargs_text['color'] = 'white'
    if fontsize is not None:
        kwargs_text['fontsize'] = fontsize
    ax.plot([x0, x1], [y0, y1], **kwargs)
    xtext, ytext = ax.transAxes.inverted().transform(ax.transData.transform([(x0 + x1) / 2.0, y0]))
    if label_position == 'above':
        verticalalign = 'bottom'
        ytext = ytext + yoff
    if label_position == 'below':
        verticalalign = 'top'
        ytext = ytext + yoff
    if xoff:
        xtext = xtext + xoff
    ax.text(xtext, ytext, label, horizontalalignment='center', verticalalignment=verticalalign, transform=ax.transAxes, **kwargs_text)


def maxfit(smap, width=[5, 5], mapxy=None):
    '''
    find maximum position of an image using 2D Gaussian fit
    :param smap: sunpy map
    :param width:
    :param mapxy: [mapx,mapy]
    :return:
    '''

    data = smap.data
    if mapxy:
        mapx, mapy = mapxy
    else:
        XX, YY = np.meshgrid(np.arange(smap.data.shape[1]), np.arange(smap.data.shape[0]))
        # mapx, mapy = smap.pixel_to_data(XX * u.pix, YY * u.pix)
        mesh = smap.pixel_to_world(XX * u.pix, YY * u.pix)
        mapx, mapy = mesh.Tx.value, mesh.Tx.value

    idxmax = np.where(data == np.nanmax(data))

    try:
        z = np.polyfit(x=mapx[idxmax[0][0], idxmax[1][0] - width[0]:idxmax[1][0] + width[0] + 1],
                       y=data[idxmax[0][0], idxmax[1][0] - width[0]:idxmax[1][0] + width[0] + 1], deg=2)
        xmax = -z[1] / z[0] / 2.0
        z = np.polyfit(x=mapy[idxmax[0][0] - width[1]:idxmax[0][0] + width[1] + 1, idxmax[1][0]],
                       y=data[idxmax[0][0] - width[1]:idxmax[0][0] + width[1] + 1, idxmax[1][0]], deg=2)
        ymax = -z[1] / z[0] / 2.0
        return [xmax, ymax]
    except:
        print('2D polynomial fitting failed.')
        return None


def contour1chn(vlafile, aiafile, chn=0, pol=0, x_range=[], y_range=[], levels=[0.2, 0.4, 0.6, 0.8], cmap='jet', *args, **kargs):
    aiamap = sunpy.map.Map(aiafile)
    if x_range and y_range:
        aiamap = aiamap.submap(u.Quantity(x_range * u.arcsec), u.Quantity(y_range * u.arcsec))
    ax = plt.subplot()
    aiamap.plot(axes=ax)
    if type(vlafile) is not list:
        vlafile = [vlafile]
    for idx, vf in enumerate(vlafile):
        hdulist = fits.open(vf)
        hdu = hdulist[0]
        dims = hdu.data.shape
        if len(dims) == 2:
            vladata = hdu.data
        elif len(dims) == 4:
            vladata = hdu.data[pol, chn, :, :]
        else:
            raise ValueError('check you import vla fits file')
        vlamap = sunpy.map.Map((vladata, hdu.header))
        sy, sx = vlamap.data.shape
        XX, YY = np.meshgrid(np.arange(sy), np.arange(sx))
        vlamapx = vlamap.pixel_to_data(XX * u.pix, YY * u.pix).Tx
        vlamapy = vlamap.pixel_to_data(XX * u.pix, YY * u.pix).Ty
        ax.contour(vlamapx.value, vlamapy.value, vlamap.data, levels=np.array(levels) * np.nanmax(vlamap.data),
                   colors=[cm.get_cmap(cmap)(float(idx) / (len(vlafile) - 1))] * len(levels), *args, **kargs)


def plot_compsite_map(vlafile, aiafile, outfile='', label='', pol=0, chans=[], chan_mask=None, x_range=[], y_range=[], levels=[0.9],
                      plotstyle='centroid', figsize=(10, 8), figdpi=100, width=[5, 5], zorder=1, maponly=False, vlaonly=False, dspecdata={},
                      thrshd=None, cmap='jet', plt_clbar=True, aiaplt_args={'axes': None, 'cmap': None, 'vmax': None, 'vmin': None}, **kwargs):
    cmap = cm.get_cmap(cmap)
    if aiaplt_args['cmap'] is None:
        aiaplt_args.pop('cmap', None)
    if outfile:
        plt.ioff()
    if not vlaonly:
        if type(aiafile) != sunpy.map.sources.sdo.AIAMap and type(aiafile) != sunpy.map.compositemap.CompositeMap:
            aiamap = sunpy.map.Map(aiafile)
        else:
            aiamap = aiafile
        if x_range and y_range:
            try:
                aiamap = aiamap.submap(u.Quantity(x_range * u.arcsec), u.Quantity(y_range * u.arcsec))
            except:
                from astropy.coordinates import SkyCoord
                bl = SkyCoord(x_range[0] * u.arcsec, y_range[0] * u.arcsec, frame=aiamap.coordinate_frame)
                tr = SkyCoord(x_range[1] * u.arcsec, y_range[1] * u.arcsec, frame=aiamap.coordinate_frame)
                aiamap = aiamap.submap(bl, tr)

    if type(vlafile) == dict:
        if dspecdata:
            if dspecdata['stack'] == 'Vstack':
                fig = plt.figure(figsize=(9, 12))
                plt.subplot(2, 1, 1)
            elif dspecdata['stack'] == 'Hstack':
                fig = plt.figure(figsize=(20, 8))
                plt.subplot(1, 2, 1)
        else:
            if aiaplt_args['axes'] is None:
                fig = plt.figure(figsize=figsize)
                plt.subplot()
            else:
                fig = plt.gcf()
        if not vlaonly:
            aiamap.plot(**aiaplt_args)
            ax1 = plt.gca()
            ax1.set_autoscale_on(False)
        else:
            ax1 = aiaplt_args['axes']
            ax1.set_autoscale_on(False)

        if label:
            ax1.text(0.98, 0.98, label, horizontalalignment='right', verticalalignment='top', color='white', transform=ax1.transAxes, fontsize=14)
        clrange = vlafile['ColorMapper']['crange']
        if not maponly:
            cargsort = np.argsort(np.array(vlafile['ColorMapper']['c']))
            cargsort = cargsort[::zorder]
            vlafile['x'] = np.array(vlafile['x'])[cargsort]
            vlafile['y'] = np.array(vlafile['y'])[cargsort]
            vlafile['ColorMapper']['c'] = np.array(vlafile['ColorMapper']['c'])[cargsort]
            vlafile['s'] = np.array(vlafile['s'])[cargsort]
            im1 = ax1.scatter(vlafile['x'], vlafile['y'], c=vlafile['ColorMapper']['c'], s=vlafile['s'], vmin=clrange[0], vmax=clrange[-1], cmap=cmap,
                              **kwargs)
        else:
            im1 = ax1.scatter([], [], c=[], s=[], vmin=clrange[0], vmax=clrange[-1], cmap=cmap, **kwargs)
        if not dspecdata:
            if plt_clbar:
                cb1 = plt.colorbar(im1, orientation='vertical', ax=ax1)
                if type(vlafile) == dict:
                    if 'title' in vlafile['ColorMapper'].keys():
                        cb1.set_label(vlafile['ColorMapper']['title'])
                    else:
                        cb1.set_label('Frequency [GHz]')
                else:
                    cb1.set_label('Frequency [GHz]')
        else:
            tim = dspecdata['time']
            dt = np.mean(np.diff(tim))
            cmapspec = cmap
            cmapspec.set_bad('white', 1.0)
            normspec = colors.Normalize(vmin=dspecdata['drange'][0], vmax=dspecdata['drange'][1])
            if dspecdata['stack'] == 'Vstack':
                plt.subplot(2, 1, 2)
            elif dspecdata['stack'] == 'Hstack':
                plt.subplot(1, 2, 2)
            ax2 = plt.gca()
            im2 = plt.pcolormesh(tim, dspecdata['freq'], dspecdata['peak'], cmap=cmapspec, norm=normspec)
            ax2.add_patch(patches.Rectangle((vlafile['t'], dspecdata['freq'][0]), dt, dspecdata['freq'][-1] - dspecdata['freq'][0], facecolor='black',
                                            edgecolor='white', alpha=0.3))
            ax2.set_xlim(tim[0], tim[-1])
            ax2.set_ylim(dspecdata['freq'][0], dspecdata['freq'][-1])
            ax2.set_title('Vector Dynamic spectrum')
            labels = ax2.get_xticks().tolist()
            newlabels = [Time(lb / 24. / 3600., format='jd').iso.split(' ')[1] for lb in labels]
            ax2.set_xticklabels(newlabels, rotation=45)
            if plt_clbar:
                cb1 = plt.colorbar(im1, orientation='vertical', ax=ax1)
                if type(vlafile) == dict:
                    if 'title' in vlafile['ColorMapper'].keys():
                        cb1.set_label(vlafile['ColorMapper']['title'])
                    else:
                        cb1.set_label('Frequency [GHz]')
                else:
                    cb1.set_label('Frequency [GHz]')
                cb2 = plt.colorbar(im2, orientation='vertical', ax=ax2)
                cb2.set_label('Max intensity [Jy/beam]')
    else:
        if aiaplt_args['axes'] is None:
            fig = plt.figure(figsize=figsize)
            plt.subplot()
        else:
            fig = plt.gcf()
        if not vlaonly:
            aiamap.plot(**aiaplt_args)
            ax1 = plt.gca()
            ax1.set_autoscale_on(False)
        else:
            ax1 = aiaplt_args['axes']
            ax1.set_autoscale_on(False)
        # ax1.xaxis.set_ticks_position('top')
        # ax1.yaxis.set_ticks_position('right')
        if label:
            ax1.text(0.98, 0.98, label, horizontalalignment='right', verticalalignment='top', color='white', transform=ax1.transAxes, fontsize=14)
        # # Prevent the image from being re-scaled while overplotting.
        hdulist = fits.open(vlafile)
        hdu = hdulist[0]
        nfreq = hdu.data.shape[1]
        if not maponly:
            vladata = hdu.data[pol, 0, :, :]
            vlamap = sunpy.map.Map((vladata, hdu.header))
            XX, YY = np.meshgrid(np.arange(vlamap.data.shape[1]), np.arange(vlamap.data.shape[0]))
            mapxy = vlamap.pixel_to_data(XX * u.pix, YY * u.pix)
            mapx, mapy = mapxy.Tx.value, mapxy.Ty.value
            clrange = (hdu.header['CRVAL3'] + np.arange(nfreq) * hdu.header['CDELT3']) / 1e9
            if len(chans) == 0:
                chans = range(0, nfreq)
            if plotstyle == 'centroid':
                pltdata = {'x': [], 'y': [], 's': [], 'c': []}
                for idx, chan in enumerate(chans):
                    vladata = hdu.data[pol, chan, :, :]
                    if np.nanmax(vladata):
                        vlamap.data = vladata
                        maxxy = maxfit(vlamap, mapxy=[mapx, mapy], width=width)
                        if maxxy:
                            x, y = maxxy
                            pltdata['x'].append(x)
                            pltdata['y'].append(y)
                            pltdata['c'].append((hdu.header['CRVAL3'] + chan * hdu.header['CDELT3']) / 1e9)
                            pltdata['s'].append(np.nanmax(vladata))
                cargsort = np.argsort(np.array(pltdata['c']))
                cargsort = cargsort[::zorder]
                pltdata['x'] = np.array(pltdata['x'])[cargsort]
                pltdata['y'] = np.array(pltdata['y'])[cargsort]
                pltdata['c'] = np.array(pltdata['c'])[cargsort]
                pltdata['s'] = np.array(pltdata['s'])[cargsort]
                im1 = ax1.scatter(pltdata['x'], pltdata['y'], c=pltdata['c'], s=pltdata['s'], vmin=clrange[0], vmax=clrange[-1], cmap=cmap, **kwargs)
                if plt_clbar:
                    cb1 = plt.colorbar(im1, orientation='vertical', ax=ax1)
                    cb1.set_label('Frequency [GHz]')
            elif plotstyle == 'contour':
                nchan = len(chans)
                for idx, chan in enumerate(chans):
                    if chan_mask is not None:
                        freq = (hdu.header['CRVAL3'] + chan * hdu.header['CDELT3']) / 1e9
                        if freq not in chan_mask:
                            continue
                    vladata = hdu.data[pol, chan, :, :]
                    vlamap = sunpy.map.Map((vladata, hdu.header))
                    SRC_vlamap_contour = DButil.get_contour_data(mapx, mapy, vlamap.data, levels=levels)
                    if SRC_vlamap_contour.data['xs']:
                        for ii, xs in enumerate(SRC_vlamap_contour.data['xs']):
                            x, y = xs, SRC_vlamap_contour.data['ys'][ii]
                            if not thrshd or np.nanmax(vladata) >= thrshd:
                                plt.plot(x, y, color=cmap(int(float(chan) / nfreq * 255)), zorder=nchan + zorder * idx, **kwargs)
                if plt_clbar:
                    fig.subplots_adjust(right=0.8)
                    cax1 = fig.add_axes([0.85, 0.1, 0.01, 0.8])
                    norm = colors.Normalize(vmin=clrange[0], vmax=clrange[-1])
                    cb1 = colorbar.ColorbarBase(cax1, cmap=cmap, norm=norm, orientation='vertical')
                    cb1.set_label('Frequency [GHz]')
    try:
        cb1.ax.set_aspect(40)
        cb1.ax.tick_params(direction='in')
    except:
        pass
    try:
        cb2.ax.set_aspect(40)
        cb2.ax.tick_params(direction='in')
    except:
        pass

    if aiaplt_args['axes'] is None:
        fig.tight_layout(pad=3)
    if outfile:
        if outfile.endswith('.eps'):
            fig.savefig(outfile, format='eps', dpi=int(figdpi))
        elif outfile.endswith('.pdf'):
            fig.savefig(outfile, format='pdf', dpi=int(figdpi))
        elif outfile.endswith('.png'):
            fig.savefig(outfile, format='png', dpi=int(figdpi))
        elif outfile.endswith('.jpeg'):
            fig.savefig(outfile, format='jpeg', dpi=int(figdpi))
        elif outfile.endswith('.jpg'):
            fig.savefig(outfile, format='jpg', dpi=int(figdpi))
        else:
            raise ValueError('Can not save {}. Provide a filename with extension of eps, pdf, png or jpeg.'.format(outfile))



from __future__ import division
import pycwt as wavelet
from pycwt.helpers import find
import numpy as np
import matplotlib.pyplot as plt


def plot_wavelet(t, dat, dt, pl, pr, period_pltlim=None, stscale=2, siglev=0.95, title='', label='', units='', tunits='', sav_img=False):
    '''

    :param t:
    :param dat:
    :param dt:
    :param pl: left edge of the period for calculating the averaged power
    :param pr: right edge of the period for calculating the averaged power
    :param title:
    :param label:
    :param units:
    :param tunits:
    :return:
    '''

    """
    In this example we will load the NINO3 sea surface temperature anomaly dataset
    between 1871 and 1996. This and other sample data files are kindly provided by
    C. Torrence and G. Compo at
    <http://paos.colorado.edu/research/wavelets/software.html>.

    """
    # We begin by importing the relevant libraries. Please make sure that PyCWT is
    # properly installed in your system.

    # Then, we load the dataset and define some data related parameters. In this
    # case, the first 19 lines of the data file contain meta-data, that we ignore,
    # since we set them manually (*i.e.* title, units).

    t0 = t[0]

    # We also create a time array in years.
    N = dat.size
    t = np.arange(0, N) * dt + t0

    # We write the following code to detrend and normalize the input data by its
    # standard deviation. Sometimes detrending is not necessary and simply
    # removing the mean value is good enough. However, if your dataset has a well
    # defined trend, such as the Mauna Loa CO\ :sub:`2` dataset available in the
    # above mentioned website, it is strongly advised to perform detrending.
    # Here, we fit a one-degree polynomial function and then subtract it from the
    # original data.
    p = np.polyfit(t - t0, dat, 1)
    dat_notrend = dat - np.polyval(p, t - t0)
    std = dat_notrend.std()  # Standard deviation
    var = std ** 2  # Variance
    dat_norm = dat_notrend / std  # Normalized dataset

    # The next step is to define some parameters of our wavelet analysis. We
    # select the mother wavelet, in this case the Morlet wavelet with
    # :math:`\omega_0=6`.
    mother = wavelet.Morlet(6)
    s0 = stscale * dt  # Starting scale, in this case 2 * 0.25 years = 6 months
    dj = 1 / 12  # Twelve sub-octaves per octaves
    J = -1  # 7 / dj  # Seven powers of two with dj sub-octaves
    alpha, _, _ = wavelet.ar1(dat)  # Lag-1 autocorrelation for red noise

    # The following routines perform the wavelet transform and inverse wavelet
    # transform using the parameters defined above. Since we have normalized our
    # input time-series, we multiply the inverse transform by the standard
    # deviation.
    wave, scales, freqs, coi, fft, fftfreqs = wavelet.cwt(dat_norm, dt, dj, s0, J, mother)
    iwave = wavelet.icwt(wave, scales, dt, dj, mother) * std

    # We calculate the normalized wavelet and Fourier power spectra, as well as
    # the Fourier equivalent periods for each wavelet scale.
    power = (np.abs(wave)) ** 2
    fft_power = np.abs(fft) ** 2
    period = 1 / freqs

    # We could stop at this point and plot our results. However we are also
    # interested in the power spectra significance test. The power is significant
    # where the ratio ``power / sig95 > 1``.
    signif, fft_theor = wavelet.significance(1.0, dt, scales, 0, alpha, significance_level=siglev, wavelet=mother)
    sig95 = np.ones([1, N]) * signif[:, None]
    sig95 = power / sig95

    # Then, we calculate the global wavelet spectrum and determine its
    # significance level.
    glbl_power = power.mean(axis=1)
    dof = N - scales  # Correction for padding at edges
    glbl_signif, tmp = wavelet.significance(var, dt, scales, 1, alpha, significance_level=siglev, dof=dof, wavelet=mother)

    # We also calculate the scale average between 2 years and 8 years, and its
    # significance level.
    sel = find((period >= pl) & (period < pr))
    Cdelta = mother.cdelta
    scale_avg = (scales * np.ones((N, 1))).transpose()
    scale_avg = power / scale_avg  # As in Torrence and Compo (1998) equation 24
    scale_avg = var * dj * dt / Cdelta * scale_avg[sel, :].sum(axis=0)
    scale_avg_signif, tmp = wavelet.significance(var, dt, scales, 2, alpha, significance_level=siglev, dof=[scales[sel[0]], scales[sel[-1]]],
                                                 wavelet=mother)

    # Finally, we plot our results in four different subplots containing the
    # (i) original series anomaly and the inverse wavelet transform; (ii) the
    # wavelet power spectrum (iii) the global wavelet and Fourier spectra ; and
    # (iv) the range averaged wavelet spectrum. In all sub-plots the significance
    # levels are either included as dotted lines or as filled contour lines.

    # Prepare the figure
    # plt.close('all')
    # plt.ioff()
    figprops = dict(figsize=(11, 8), dpi=72)
    fig = plt.figure(**figprops)

    # First sub-plot, the original time series anomaly and inverse wavelet
    # transform.
    ax = plt.axes([0.1, 0.75, 0.65, 0.2])
    ax.plot(t, iwave, '-', linewidth=1, color=[0.5, 0.5, 0.5])
    ax.plot(t, dat, 'k', linewidth=1.5)
    ax.set_title('a) {}'.format(title))
    ax.set_ylabel(r'{} [{}]'.format(label, units))

    # Second sub-plot, the normalized wavelet power spectrum and significance
    # level contour lines and cone of influece hatched area. Note that period
    # scale is logarithmic.
    bx = plt.axes([0.1, 0.37, 0.65, 0.28], sharex=ax)
    levels = [0.0625, 0.125, 0.25, 0.5, 1, 2, 4, 8, 16]
    bx.contourf(t, np.log2(period), np.log2(power), np.log2(levels), extend='both', cmap=plt.cm.viridis)
    extent = [t.min(), t.max(), 0, max(period)]
    bx.contour(t, np.log2(period), sig95, [-99, 1], colors='k', linewidths=2, extent=extent)
    bx.fill(np.concatenate([t, t[-1:] + dt, t[-1:] + dt, t[:1] - dt, t[:1] - dt]),
            np.concatenate([np.log2(coi), [1e-9], np.log2(period[-1:]), np.log2(period[-1:]), [1e-9]]), 'k', alpha=0.3, hatch='x')
    bx.set_title('b) {} Wavelet Power Spectrum ({})'.format(label, mother.name))
    bx.set_ylabel('Period ({})'.format(tunits))
    #
    Yticks = 2 ** np.arange(np.ceil(np.log2(period.min())), np.ceil(np.log2(period.max())))
    bx.set_yticks(np.log2(Yticks))
    bx.set_yticklabels(Yticks)

    # Third sub-plot, the global wavelet and Fourier power spectra and theoretical
    # noise spectra. Note that period scale is logarithmic.
    cx = plt.axes([0.77, 0.37, 0.2, 0.28], sharey=bx)
    cx.plot(glbl_signif, np.log2(period), 'k--')
    cx.plot(var * fft_theor, np.log2(period), '--', color='#cccccc')
    cx.plot(var * fft_power, np.log2(1. / fftfreqs), '-', color='#cccccc', linewidth=1.)
    cx.plot(var * glbl_power, np.log2(period), 'k-', linewidth=1.5)
    cx.set_title('c) Global Wavelet Spectrum')
    cx.set_xlabel(r'Power [({})^2]'.format(units))
    cx.set_xlim([0, np.nanmax([np.nanmax(glbl_power * var), np.nanmax(var * fft_power)])])
    if period_pltlim:
        cx.set_ylim(np.log2(period_pltlim))
    else:
        cx.set_ylim(np.log2([period.min(), period.max()]))
    cx.set_yticks(np.log2(Yticks))
    cx.set_yticklabels(Yticks)
    plt.setp(cx.get_yticklabels(), visible=False)

    # Fourth sub-plot, the scale averaged wavelet spectrum.
    dx = plt.axes([0.1, 0.07, 0.65, 0.2], sharex=ax)
    dx.axhline(scale_avg_signif, color='k', linestyle='--', linewidth=1.)
    dx.plot(t, scale_avg, 'k-', linewidth=1.5)
    dx.set_title('d) {}--{} {} scale-averaged power'.format(pl, pr, tunits))
    dx.set_xlabel('Time ({})'.format(tunits))
    dx.set_ylabel(r'Average variance [{}]'.format(units))
    ax.set_xlim([t.min(), t.max()])

    plt.show()

    if sav_img:
        if type(sav_img) is str:
            fig.savefig(sav_img)
        else:
            fig.savefig('wavelet.png')



import sys, os, traceback, glob
from time import sleep
from PyQt5.QtWidgets import QMainWindow, QFileDialog, QApplication, QPushButton, QCheckBox, QLineEdit, QLabel, QTextEdit, QListWidgetItem, QWidget, QAction, QTabWidget, QTableView, QVBoxLayout, QHBoxLayout, QGridLayout, QStatusBar, QProgressBar, QFrame, QFormLayout, QGroupBox, QButtonGroup, QRadioButton, QComboBox, QSizePolicy

from PyQt5.QtGui import QIcon, QFont, QColor
from PyQt5.QtCore import pyqtSlot, Qt, QAbstractTableModel, pyqtSignal, QRunnable, QObject, QThreadPool
from pr_summary import pr_summary
from casatools import ms as mstool

from astropy.time import Time
from casatools import table
from astropy.io import fits
from copy import deepcopy
from functools import partial
from matplotlib.backends.backend_qt5agg import (
        FigureCanvas, NavigationToolbar2QT as NavigationToolbar)
from matplotlib.figure import Figure
from matplotlib.dates import DateFormatter
import numpy as np


class WorkerSignals(QObject):
    '''
    Defines the finished signal available from a running worker thread.
    '''
    finished = pyqtSignal()
    error = pyqtSignal(tuple)


class Worker(QRunnable):
    '''
    Worker thread

    Inherits from QRunnable to handler worker thread setup, signals and wrap-up.

    :param callback: The function callback to run on this worker thread. Supplied args and
                     kwargs will be passed through to the runner.
    :type callback: function
    :param args: Arguments to pass to the callback function
    :param kwargs: Keywords to pass to the callback function

    '''

    def __init__(self, fn, *args, **kwargs):
        super(Worker, self).__init__()

        # Store constructor arguments (re-used for processing)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.signals = WorkerSignals()

        # Add the callback to our kwargs
        #self.kwargs['progress_callback'] = self.signals.progress

    @pyqtSlot()
    def run(self):
        '''
        Initialise the runner function with passed args, kwargs.
        '''

        # Retrieve args/kwargs here; and fire processing using them
        try:
            #print(self.fn)
            #print(*self.args)
            result = self.fn(*self.args, **self.kwargs)
        except:
            traceback.print_exc()
            exctype, value = sys.exc_info()[:2]
            self.signals.error.emit((exctype, value, traceback.format_exc()))
        finally:
            self.signals.finished.emit()  # Done

class TableModel(QAbstractTableModel):
    ''' Defines the action of the tclean table
    '''
    def __init__(self, data):
        super(TableModel, self).__init__()
        self._data = data
        
    def data(self, index, role):
        if role == Qt.DisplayRole:
            if index.column() > 1: return ""
            return self._data[index.row()][index.column()]
        if role == Qt.BackgroundRole:
            if self._data[index.row()][0][0] != ' ':
                return QColor('#ffccaa')
        if role == Qt.EditRole:
            return self._data[index.row()][index.column()]
        if role == Qt.ToolTipRole:
            if index.column() == 0:
                if len(self._data[index.row()]) == 4:
                    return self._data[index.row()][3]
            
    def setData(self, index, value, role):
        if role == Qt.EditRole:
            #value = self.checkdata(index, value)
            self._data[index.row()][index.column()] = value
            self.dataChanged.emit(index, index)
            return True
            
    def rowCount(self, index):
        return len(self._data)
        
    def columnCount(self, index):
        return len(self._data[0])

    def flags(self, index):
        if index.column() !=3:
            flag_item = Qt.ItemIsSelectable|Qt.ItemIsEnabled
        if index.column() == 1 and len(self._data[index.row()][2]) == 0: 
            flag_item |= Qt.ItemIsEditable
        return flag_item 

class App(QMainWindow):

    def __init__(self):
        super().__init__()
        self.title = 'EOVSA Imager'
        self.left = 0
        self.top = 0
        self.width = 1200
        self.height = 900
        self.setWindowTitle(self.title)
        self.setGeometry(self.left, self.top, self.width, self.height)
        self._main = QWidget()
        self.setCentralWidget(self._main)
        self.initUI()
        self.threadpool = QThreadPool()


    def initUI(self):

        self.statusBar = QStatusBar()
        self.setStatusBar(self.statusBar)
        self.progressBar = QProgressBar()
        self.progressBar.setGeometry(10,10,200,15)
        
        layout = QVBoxLayout()
        # Initialize tab screen
        self.tabs = QTabWidget()
        tab1 = QWidget()
        tab2 = QWidget()
        tab3 = QWidget()
        tab4 = QWidget()
        tab5 = QWidget()
        tab6 = QWidget()

        # Add tabs
        self.tabs.addTab(tab1,"Data Select")
        self.tabs.addTab(tab2,"Viewer")
        self.tabs.addTab(tab3,"Imager")
        self.tabs.addTab(tab4,"Selfcal")
        self.tabs.addTab(tab5,"Production")
        self.tabs.addTab(tab6,"Export")

        # Each tab's user interface is complex, so this splits them into separate functions.
        self.initUItab1()
        self.initUItab2()
        self.initUItab3()
        self.initUItab4()
        self.initUItab5()
        self.initUItab6()

        self.tabs.currentChanged.connect(self.tabChanged)
        
        # Add tabs to widget
        layout.addWidget(self.tabs)
        self._main.setLayout(layout)

        self.show()

    def tabChanged(self, i):
        if i == 2:
            self.update_params()

#
# Data Select Tab User Interface        
#
    def initUItab1(self):
        # Create main layout (a Vertical Layout)
        mainlayout = QVBoxLayout()

        # Create Data Select tab
        upperbox = QHBoxLayout()  # Has two hboxes: leftbox and msinfobox
        leftbox = QVBoxLayout()   # Has a gridlayout: filenamebox; and groupBox: selectBox
        # Filename entry
        filenamebox = QGridLayout()
        leftbox.addLayout(filenamebox)
        # Create LineEdit widget for ms filename
        self.msentry = QLineEdit()
        self.fname = '<Select or enter a valid ms filename>'
        self.msentry.resize(8*len(self.fname),20)
        self.ms = None       # No ms yet
        self.msdata = None  # No amplitude data read from ms yet
        self.msentry.setText(self.fname)
        self.msentry.returnPressed.connect(self.on_return)
        filenamebox.addWidget(QLabel("MS Filename"),0,0)
        filenamebox.addWidget(self.msentry,1,0,1,4)
        # Create Browse button
        myButton1 = QPushButton("Browse")
        myButton1.clicked.connect(self.on_click)
        filenamebox.addWidget(myButton1,1,5)
        upperbox.addLayout(leftbox)

        # Create label and TextEdit widget for ms information
        msinfobox = QVBoxLayout()
        upperbox.addLayout(msinfobox)
        self.infoEdit = QTextEdit()
        #f = QFont("Courier",9)
        #self.infoEdit.setCurrentFont(f)
        self.infoEdit.setReadOnly(True)
        self.infoEdit.setMinimumHeight(300)
        self.infoEdit.setMinimumWidth(550)
        msinfobox.addWidget(QLabel("MS Information"))
        msinfobox.addWidget(self.infoEdit)
        mainlayout.addLayout(upperbox)

        # Data Selection
        selectBox = QGroupBox("Data Selection Criteria")
        selectarea = QFormLayout()
        self.trangeEdit = QLineEdit()
        selectarea.addRow("Timerange", self.trangeEdit)
        self.spwEdit = QLineEdit()
        selectarea.addRow("Sp. Window", self.spwEdit)
        self.baselineEdit = QLineEdit()
        selectarea.addRow("Baseline", self.baselineEdit)
        self.stokesEdit = QLineEdit()
        selectarea.addRow("Stokes", self.stokesEdit)
        self.uvrangeEdit = QLineEdit()
        selectarea.addRow("UV range", self.uvrangeEdit)
        selectBox.setLayout(selectarea)
        leftbox.addWidget(selectBox)
        drawselect = QPushButton('Draw Selection on Plot')
        drawselect.clicked.connect(self.drawsel)
        leftbox.addWidget(drawselect)
        leftbox.addStretch(1)
        
        #hbox.addStretch(1)
        
        playarea = QHBoxLayout()
        playarea.addStretch(1)
        xferarea = QVBoxLayout()
        xferDownRight = QPushButton('Use Selection for Plot Limits')
        xferDownRight.setIcon(QIcon('icons/down_right.png'))
        xferDownRight.clicked.connect(self.xferDR)
        xferarea.addWidget(xferDownRight)
        xferUpLeft = QPushButton('Use Plot Limits for Selection')
        xferUpLeft.setIcon(QIcon('icons/up_left.png'))
        xferUpLeft.clicked.connect(self.xferUL)
        xferarea.addWidget(xferUpLeft)
        xferarea.addStretch(1)
        playarea.addLayout(xferarea)
        # Add a figure to the canvas.
        plotarea = QVBoxLayout()
        self.speccanvas = FigureCanvas(Figure(figsize=(8, 3)))
        plotarea.addWidget(self.speccanvas)
        xferarea.addWidget(NavigationToolbar(self.speccanvas, self))
        maxmin = QHBoxLayout()
        minlabel = QLabel('Min[sfu]')
        maxlabel = QLabel('Max[sfu]')
        self.minentry = QLineEdit()
        self.maxentry = QLineEdit()
        maxmin.addStretch(1)
        maxmin.addWidget(minlabel)
        maxmin.addWidget(self.minentry)
        maxmin.addWidget(maxlabel)
        maxmin.addWidget(self.maxentry)
        maxmin.addStretch(1)
        self.ignore_gaps = QCheckBox('Ignore Gaps')
        maxmin.addWidget(self.ignore_gaps)
        maxmin.addStretch(1)
        self.minentry.returnPressed.connect(self.plot_data)
        self.maxentry.returnPressed.connect(self.plot_data)
        plotarea.addLayout(maxmin)
        playarea.addLayout(plotarea)
#        self.addToolBar(Qt.BottomToolBarArea, NavigationToolbar(self.speccanvas, self))
        self.dspec_ax = self.speccanvas.figure.subplots()
        baseline_layout = QGridLayout()
        baseline_layout.setSpacing(1)
        ant1 = []
        ant2 = []
        nant = 13
        bl2ord = get_bl2ord(nant)     # Lookup table for 13-ant baseline matrix
        nbl = int(nant*(nant+1)/2)
        self.blcheck = [0]*nbl        # This will hold the check box widgets for the baselines
        self.blchecked = [False]*nbl  # This is the state of the buttons
        self.prev = None
        for i in range(nant):
            ant1.append(QPushButton('{:2d}'.format(i+1)))
            ant1[-1].setCheckable(True)
            ant1[-1].setMaximumSize(12,12)   # Ant button size in pixels
            ant1[-1].setStyleSheet("font : 8px;")
            ant1[-1].clicked.connect(partial(self.ant_click,ant1[-1]))
            baseline_layout.addWidget(ant1[-1],0,i)    # Grid location of ant buttons along top
            ant2.append(QPushButton('{:2d}'.format(i+1)))
            ant2[-1].setCheckable(True)
            ant2[-1].setMaximumSize(12,12)   # Ant button size in pixels
            ant2[-1].setStyleSheet("font : 8px;")
            ant2[-1].clicked.connect(partial(self.ant_click,ant2[-1]))
            baseline_layout.addWidget(ant2[-1],i+1,nant)  # Grid location of ant buttons along side
            for j in range(i,nant):
                if i == j:
                    self.blcheck[bl2ord[i,j]] = QCheckBox("")
                    button = self.blcheck[bl2ord[i,j]]  # Just saves typing
                    button.setStyleSheet("background-color : #8888ff;")
                else:
                    self.blcheck[bl2ord[i,j]] = QCheckBox("")
                    button = self.blcheck[bl2ord[i,j]]  # Just saves typing
                    button.setStyleSheet("background-color : #ff8888;")
                button.setMaximumSize(12,12)
                button.clicked.connect(self.baselineClicked)
                baseline_layout.addWidget(button,i+1,j)   # Grid location of baseline buttons
        self.ant1buttons = ant1
        self.ant2buttons = ant2
        unsel = QPushButton('Unselect All')
        unsel.setStyleSheet("font : 10px;")
        unsel.clicked.connect(partial(self.sel_clicked, False))
        sel = QPushButton('Select All')
        sel.setStyleSheet("font : 10px;")
        sel.clicked.connect(partial(self.sel_clicked, True))
        baseline_layout.addWidget(unsel,9,0,2,7)
        baseline_layout.addWidget(sel,11,0,2,7)
        space = QVBoxLayout()
        # Add Polarization Button Group
        self.polGroup = QButtonGroup()
        self.polGroup.buttonClicked.connect(self.get_data)
        xxButton = QRadioButton("XX")
        xxButton.setChecked(True)
        yyButton = QRadioButton("YY")
        xyButton = QRadioButton("XY")
        yxButton = QRadioButton("YX")
        rrButton = QRadioButton("RR")
        llButton = QRadioButton("LL")
        self.polGroup.addButton(xxButton)
        self.polGroup.addButton(yyButton)
        self.polGroup.addButton(xyButton)
        self.polGroup.addButton(yxButton)
        self.polGroup.addButton(rrButton)
        self.polGroup.addButton(llButton)
        pspace = QHBoxLayout()
        pspace.addWidget(xxButton)
        pspace.addWidget(yyButton)
        pspace.addWidget(xyButton)
        pspace.addWidget(yxButton)
        pspace.addWidget(rrButton)
        pspace.addWidget(llButton)
        space.addStretch(1)
        space.addLayout(pspace)
        space.addWidget(QLabel('Baseline Selection Map'))
        space.addLayout(baseline_layout)
        space.addStretch(1)
        gobutton = QPushButton('Go')
        gobutton.clicked.connect(self.get_data)
        gobutton.setStyleSheet("background-color : #ffff88;")
        space.addWidget(gobutton)
        playarea.addLayout(space)
        mainlayout.addLayout(playarea)
        self.tabs.widget(0).setLayout(mainlayout)

#
# Viewer Tab User Interface        
#
    def initUItab2(self):
        # Create main layout (a Vertical Layout)
        mainlayout = QVBoxLayout()
        mainlayout.addWidget(QLabel("This is the viewer tab"))
        self.tabs.widget(1).setLayout(mainlayout)

#
# Imager Tab User Interface        
#
    def initUItab3(self):
        # Create main layout (a Horizontal Layout)
        mainlayout = QHBoxLayout()
        # Create a left and right side (both Vertical Layouts)
        leftlayout = QVBoxLayout()
        rightlayout = QVBoxLayout()
        
        # Create a table for interacting with parameters
        self.table = QTableView()
        self.table.doubleClicked.connect(self.tblRowClicked)
        # For playing with, I will make a static table of tclean parameters
        tbl = [
                  ["Data Selection","",[]],
                  ["  vis","''",[]],
                  #["  field","''",[]],
                  ["  spw","''",[],"Spectral Window:\n default: ''=all; examples:\n spw='0~2,4'; spectral windows 0,1,2,4 (all channels)\n spw='0:5~61'; spw 0, channels 5 to 61\n spw='<2';   spectral windows less than 2 (i.e. 0,1)" ],
                  ["  timerange","''",[],"Range of time to select from data\n default: '' (all); examples:\n timerange = 'YYYY/MM/DD/hh:mm:ss~YYYY/MM/DD/hh:mm:ss'\n Note: if YYYY/MM/DD is missing date defaults to first day in data set\n timerange='09:14:0~09:54:0' picks 40 min on first day\n timerange='25:00:00~27:30:00' picks 1 hr to 3 hr 30 min on NEXT day\n timerange='09:44:00' pick data within one integration of time\n timerange='> 10:24:00' data after this time"],
                  ["  uvrange","''",[],"Select data within uvrange (default unit is meters) [default: '' (all)]\n examples:\n uvrange='0~1000klambda'; uvrange from 0-1000 kilo-lambda\n uvrange='> 4klambda';uvranges greater than 4 kilo lambda"],
                  ["  antenna","''",[],"Select data on 0-based antenna/baseline index [default: '' (all)]\n examples:\n antenna='0~5,7~12&0~5,7~12'; all baselines not including antenna index 6\n antenna='5&6;7&8'; baselines 5-6 and 7-8\n antenna='5'; all baselines with antenna index 5\n antenna='5,6,9'; all baselines with antenna index numbers 5,6,9"],
                  ["  datacolumn","'data'",["'data'","'corrected'","'model'"]],
                  ["Image Definition","",[]],
                  ["  imagename","''",[]],
                  ["  imsize","[128,128]",[]],
                  ["  cellsize","'2arcsec'",[]],
                  ["  phaseshift","[0, 0]",[],"X, Y offset of center of map from Sun Center"],
                  ["  stokes","'XX'",["'XX'","'YY'","'I'","'V'","'IV'","'RR'","'LL'"]],
                  ["  startmodel", "''",[]],
                  ["  specmode","'mfs'",["'mfs'","'cubedata'"]],
                  ["Deconvolution Options","",[]],
                  ["  deconvolver","'multiscale'",["'hogbom'","'clark'","'clarkstokes'","'multiscale'","'mem'"]],
                  ["  scales", "[1,5,10]",[]],
                  ["  restoringbeam","''",[]],
                  ["  pbcor","False",['True','False']],
                  ["Weighting","",[]],
                  ["  weighting","'briggs'",["'natural'","'uniform'","'briggs'"]],
                  ["  robust","0.5",[]],
                  ["  uvtaper","''",[]],
                  ["Other Options","",[]],
                  ["  niter","0",[]],
                  ["  gain","0.1",[]],
                  ["  threshold","0",[]],
                  ["  interactive","False",['True','False']],
                  ["  mask","''",[]]]
        
        self.table.verticalHeader().setDefaultSectionSize(14)  # Sets height of cells to 14px
        self.table.setModel(TableModel(tbl))  # The TableModel class is defined at the top of this file
        
        # For parameters that have to be only a limited set of fixed values, create a combobox dropdown
        # for editing them.
        for idx in range(len(tbl)):
            if len(tbl[idx]) > 2:
                if len(tbl[idx][2]) > 1:
                    i = self.table.model().index(idx,2)
                    c = QComboBox()
                    for item in tbl[idx][2]:
                        c.addItem(item)
                    c.currentTextChanged.connect(self.handleCombo)
                    c.index = self.table.model().index(idx,1)
                    self.table.setIndexWidget(i,c)

        self.table.resizeColumnsToContents()
        self.table.model().dataChanged.connect(self.update_view)

        # Determine the header rows in the table (indicated by NOT starting with a blank space).
        self.headerrows = []
        for i,tblrow in enumerate(tbl):
            if tbl[i][0][0] != ' ':
                self.headerrows.append(i)
                self.table.setSpan(i,0,1,2)
        self.headerrows.append(len(tbl))  #Add length of table, for finding length of last section
        
        titlelayout = QHBoxLayout()
        titlelayout.addWidget(QLabel("TCLEAN Parameters"))
        titlelayout.addSpacing(100)
        updateButton = QPushButton("Update Parameters")
        titlelayout.addWidget(updateButton)
        updateButton.clicked.connect(self.update_params)
        titlelayout.addSpacing(100)
        scriptButton = QPushButton("Save to CASA Script")
        titlelayout.addWidget(scriptButton)
        scriptButton.clicked.connect(self.save2CASAscript)
        titlelayout.addStretch(1)
        tablelayout = QHBoxLayout()
        self.table.setMinimumSize(600,300)
        tablelayout.addWidget(self.table)
        self.nofits = QCheckBox('Skip conversion to FITS')
        self.nocleanup = QCheckBox('Keep CASA image files')
        tablelayout.addStretch(1)
        leftlayout.addLayout(titlelayout)
        leftlayout.addLayout(tablelayout)
        leftlayout.addWidget(self.nofits)
        leftlayout.addWidget(self.nocleanup)
        self.scriptEdit = QTextEdit()
        #f = QFont("Courier",9)
        #self.infoEdit.setCurrentFont(f)
        self.scriptEdit.setReadOnly(True)
        self.scriptEdit.setMinimumHeight(300)
        self.scriptEdit.setMinimumWidth(550)
        leftlayout.addWidget(QLabel("Generated Script"))
        leftlayout.addWidget(self.scriptEdit)
        execButton = QPushButton('Execute Script')
        execButton.clicked.connect(self.execscript)
        eblayout = QHBoxLayout()
        eblayout.addWidget(execButton)
        eblayout.addStretch(1)
        leftlayout.addLayout(eblayout)
        leftlayout.addStretch(1)
        mainlayout.addLayout(leftlayout)
        
        self.imgcanvas = FigureCanvas(Figure(figsize=(7, 6)))
        rightlayout.addWidget(self.imgcanvas)
        rightlayout.addWidget(NavigationToolbar(self.imgcanvas, self))
        self.img_ax = self.imgcanvas.figure.subplots()

        mainlayout.addLayout(rightlayout)
        
        self.tabs.widget(2).setLayout(mainlayout)
        
    def save2CASAscript(self):
        script = ['from casatasks import tclean']
        tbl = self.table.model()._data
        for k,row in enumerate(tbl):
            if row[0][0] == ' ':
                # This is a parameter row.  Most such rows can be defined directly, but some require translation
                param = row[0].lstrip()
                if param == 'phaseshift':
                    # Determine appropriate phase center and set accordingly
                    try:
                        pshift = [float(i) for i in row[1].replace('[','').replace(']','').split(',')]
                    except:
                        self.statusBar.showMessage('Error translating phaseshift parameter--using zero',2000)
                        pshift = [0.0,0.0]
                    script.append("phasecenter='"+pshift2pcenter(pshift,self.ms)+"'")
                elif param == 'imagename':
                    if row[1] == "''":
                        msstem = os.path.basename(os.path.splitext(self.fname)[0])
                        script.append("imagename='images/"+msstem+"'")
                    else:
                        script.append(param+"="+row[1])
                else:
                    if row[1] == "''":
                        script.append(param+"=''")
                    else:
                        script.append(param+"="+row[1])

        script.append("tclean(vis=vis, selectdata=True, field='', spw=spw, timerange=timerange, uvrange=uvrange, antenna=antenna, scan='', observation='', intent='', datacolumn=datacolumn, imagename=imagename, imsize=imsize, cell=cellsize, phasecenter=phasecenter, stokes=stokes, projection='SIN', startmodel='', specmode=specmode, reffreq='', nchan=-1, start='', width='', outframe='LSRK', veltype='radio', restfreq=[], interpolation='linear', perchanweightdensity=True, gridder='standard', facets=1, psfphasecenter='', chanchunks=1, wprojplanes=1, vptable='', mosweight=True, aterm=True, psterm=False, wbawp=True, conjbeams=False, cfcache='', usepointing=False, computepastep=360.0, rotatepastep=360.0, pointingoffsetsigdev=[], pblimit=0.2, normtype='flatnoise', deconvolver=deconvolver, scales=scales, nterms=2, smallscalebias=0.0, restoration=True, restoringbeam=restoringbeam, pbcor=pbcor, outlierfile='', weighting=weighting, robust=robust, noise='1.0Jy', npixels=0, uvtaper=uvtaper, niter=niter, gain=gain, threshold=threshold, nsigma=0.0, cycleniter=-1, cyclefactor=1.0, minpsffraction=0.05, maxpsffraction=0.8, interactive=interactive, usemask='user', mask=mask, pbmask=0.0, sidelobethreshold=3.0, noisethreshold=5.0, lownoisethreshold=1.5, negativethreshold=0.0, smoothfactor=1.0, minbeamfrac=0.3, cutthreshold=0.01, growiterations=75, dogrowprune=True, minpercentchange=-1.0, verbose=False, fastnoise=True, restart=True, savemodel='none', calcres=True, calcpsf=True, parallel=False)")
        if not self.nofits.isChecked():
            script.append("import helioim2fits as hf")
            script.append("fitsfile = hf.imreg(vis=vis,imagefile=imagename+'.image',fitsdir='fits',timerange=timerange)")
        if not self.nocleanup.isChecked():
            script.append("import glob, shutil")
            script.append("for file in glob.glob(imagename+'.*'): shutil.rmtree(file)")
            if self.nofits.isChecked():
                self.statusBar.showMessage('Warning! You have selected no FITs output and do not keep CASA images, so you will get no output!',2000)
        self.script = script
        self.scriptEdit.setPlainText('\n'.join(script))

    def execscript(self):
        #print('Before call:')
        #print(exec)
        #print(['\n'.join(self.script)])
        worker = Worker(exec, '\n'.join(self.script))
        worker.signals.finished.connect(self.thread_complete)
        self.threadpool.start(worker)
        print('The thread is started--just waiting for the signal.')
        
    def thread_complete(self):
        print('The script execution thread is done!')
        print('The fits file should have been completed, and can be read now...')
        sleep(1) # Make sure file is closed...
        list_of_files = glob.glob('fits/*')
        latest_file = max(list_of_files, key=os.path.getctime)
        img, h = fits.getdata(latest_file, header=True)
        img.shape = (h['NAXIS2'],h['NAXIS1'])
        self.img_ax.cla()
        xval = np.linspace((h['CRVAL1']-h['NAXIS1'])*h['CDELT1']/2., (h['CRVAL1']+h['NAXIS1'])*h['CDELT1']/2.,h['NAXIS1'])
        yval = np.linspace((h['CRVAL2']-h['NAXIS2'])*h['CDELT2']/2., (h['CRVAL2']+h['NAXIS2'])*h['CDELT2']/2.,h['NAXIS2'])
        im = self.img_ax.pcolormesh(xval, yval, img)
        
        self.img_ax.set_aspect('equal')
        self.img_ax.set_ylabel('Y [arcsec]')
        self.img_ax.set_xlabel('X [arcsec]')
        self.img_ax.set_title(latest_file)
        self.imgcanvas.draw()
        
    def update_params(self):
        ''' Take entries from Data Select tab and update the tclean parameter table
        '''
        tbl = self.table.model()._data
        params = ['vis', 'spw', 'timerange', 'uvrange', 'antenna', 'stokes']
        curvalues = [self.fname, self.spwEdit.text(), self.trangeEdit.text(), self.uvrangeEdit.text(), 
                                    self.baselineEdit.text(), self.stokesEdit.text()]
        for k,row in enumerate(tbl):
            r = row[0][2:]  # parameter name (after removing two blank spaces)
            for idx, p in enumerate(params):
                if r == p:
                    i = self.table.model().index(k,1)   # Table index of parameter value 
                    # Write the new value into the table (all of these are strings, so include quotes)
                    self.table.model().setData(i, "'"+curvalues[idx]+"'", Qt.EditRole)
                    if r == 'stokes':
                        # Stokes has a combobox, so set it according to the updated parameter
                        i2 = self.table.model().index(k,2)  # Table index of combobox widget
                        c = self.table.indexWidget(i2)
                        loc = c.findText("'"+curvalues[idx]+"'")
                        if loc == -1:
                            # If curvalue is not found in the combobox, indicate stokes as unset.
                            self.table.model().setData(i, '<unset>', Qt.EditRole)
                        else: 
                            c.setCurrentIndex(loc)

    def update_view(self):
        self.table.resizeColumnsToContents()
    
    def handleCombo(self):
        sender = self.sender()
        text = sender.currentText()
        self.table.model().setData(sender.index, text, Qt.EditRole)
        
    def tblRowClicked(self, index):
        ''' If a header row is double-clicked, show or hide the rows corresponding to that section
        '''
        hrows = self.headerrows   # Row numbers corresponding to headers   
        for n,row in enumerate(hrows):
            if index.row() == row:
                # The row of the clicked cell is a header row, so identify the lines between it
                # and the next header and toggle hidden
                k1 = hrows[n] + 1
                k2 = hrows[n+1]
                if self.table.isRowHidden(k1):
                    # Rows are aleady hidden, so show them
                    for i in range(k1,k2):
                        self.table.showRow(i)
                else:
                    # Rows are visible, so hide them
                    for i in range(k1,k2):
                        self.table.hideRow(i)
                break
#     
# Selfcal Tab User Interface        
#
    def initUItab4(self):
        # Create main layout (a Vertical Layout)
        mainlayout = QVBoxLayout()
        mainlayout.addWidget(QLabel("This is the selfcal tab"))
        self.tabs.widget(3).setLayout(mainlayout)

#
# Production Tab User Interface        
#
    def initUItab5(self):
        # Create main layout (a Vertical Layout)
        mainlayout = QVBoxLayout()
        mainlayout.addWidget(QLabel("This is the production tab"))
        self.tabs.widget(4).setLayout(mainlayout)

#
# Export Tab User Interface        
#
    def initUItab6(self):
        # Create main layout (a Vertical Layout)
        mainlayout = QVBoxLayout()
        mainlayout.addWidget(QLabel("This is the export tab"))
        self.tabs.widget(5).setLayout(mainlayout)

    def baselineClicked(self):
        button = self.sender()
        if button.isChecked():
            if not self.prev is None:
                if self.prev.isChecked():
                    self.prev.setChecked(False)
            self.prev = button
            self.get_data()
        
    def on_return(self):
        ''' Called when the ms filename LineEdit widget gets a carriage-return.
            Trys to connect to the ms and return some info (no data read at this time)
        '''
        self.fname = self.msentry.text()
        self.msdata = None
        try:
            if self.ms:
                self.ms.close()
            else:
                self.ms = mstool()
            self.ms.open(self.fname)
            self.msentry.setText(self.fname)
            lines = pr_summary(self.ms)
            self.infoEdit.setPlainText('\n'.join(lines))
        except:
            self.statusBar.showMessage('Filename is not a valid ms',2000)
            self.fname = '<Select or enter a valid ms filename>'
            self.msentry.setText(self.fname)
            self.infoEdit.setPlainText('')
        try:
            t1str, t2str = lines[2].replace('-','/').split(': ')[1:3]
            trange = t1str[0:10]+'/'+t1str[11:21]+'~'+t2str[0:10]+'/'+t2str[11:21]
        except:
            trange = '<unset>'
        try:
            spwstr = lines[-1].strip().split(' ')[0]
            spw = '0~'+spwstr
        except:
            spw = '<unset>'
        self.trangeEdit.setText(trange)
        self.spwEdit.setText(spw)
        self.baselineEdit.setText('0~12&0~12')
        self.stokesEdit.setText('XX')
        self.uvrangeEdit.setText('0~2km')
        if spwstr == '30':
            self.ignore_gaps.setChecked(True)
        else:
            self.ignore_gaps.setChecked(False)

    def on_click(self):
        ''' Handle Browse button '''
        self.fname = QFileDialog.getExistingDirectory(self, 'Select MS','./',QFileDialog.ShowDirsOnly)
        self.msentry.setText(self.fname)
        self.on_return()

    def ant_click(self, button):
        ''' An antenna button was clicked, so check the checkboxes corresponding
            to the currently checked antennas.  button is a dictionary with keys
            'button' (actual button handle) and 'iant' (ordinal location in list
            of buttons).  
            Each antenna has two buttons, so we have to ensure that both are toggled.  
            iant is an integer from 0-25.  If even, toggle buttons iant and iant+1.  
            If odd, toggle buttons iant and iant-1.
        '''
        k = 0
        ant = -1
        for b in self.ant1buttons:
            if b == button:
                self.ant2buttons[k].setChecked(button.isChecked())
                ant = k
                break
            else:
                k+=1
        k = 0
        for b in self.ant2buttons:
            if b == button:
                self.ant1buttons[k].setChecked(button.isChecked())
                ant = k
                break
            else:
                k+=1
        self.statusBar.showMessage('Antenna '+str(ant+1)+' hit',1000)
        nant = 13
        bl2ord = get_bl2ord(nant)
        # Reflect antenna state on baselines
        for i in range(nant):
            for j in range(i,nant):
                if i != j:
                    if i == ant:
                        self.blcheck[bl2ord[i,j]].setChecked(button.isChecked())
                    if j == ant:
                        self.blcheck[bl2ord[i,j]].setChecked(button.isChecked())
        # Have to go back and set any baselines whose antennas are checked.
        for i in range(nant):
            if self.ant1buttons[i].isChecked():
                for k in range(i):
                    self.blcheck[bl2ord[k,i]].setChecked(True)
                for j in range(i+1,nant):
                    self.blcheck[bl2ord[i,j]].setChecked(True)
                    
    def sel_clicked(self, state):
        nant = 13
        bl2ord = get_bl2ord(nant)
        for i in range(nant):
            self.ant1buttons[i].setChecked(state)
            self.ant2buttons[i].setChecked(state)
            for j in range(i,nant):
                self.blcheck[bl2ord[i,j]].setChecked(state)

    def get_data(self):
        # Find which baselines are checked:
        bl = []
        blid = []
        isauto = []
        ord13 = get_bl2ord(13)
        ord16 = get_bl2ord(16)
        for i in range(13):
            for j in range(i,13):
                if self.blcheck[ord13[i,j]].isChecked():
                    bl.append(ord16[i,j])
                    isauto.append(i == j)
                    blid.append([i,j])
        if len(bl) == 0:
            self.statusBar.showMessage('No baselines have been selected.',2000)
            return
        if len(bl) != 1:
            self.statusBar.showMessage('Only one baseline can be shown.',2000)
            return

        # If no data have been read, read the data
        if self.msdata is None:
            self.read_msdata(fillnan=np.nan)
        npol, nf, nbl, nt = self.msdata['data'].shape
        self.plotbl = bl[0]

        # Determine which polarization state is selected by radiobuttons
        polnum = {'XX':0, 'YY':1, 'XY':2, 'YX':3, 'RR':4, 'LL':5}
        ipol = polnum[self.polGroup.checkedButton().text()]
        
        if ipol < 4:
            # The selected polarization is just one of the native polarizations (XX, YY, XY, or YX)
            spec = np.abs(self.msdata['data'][ipol,:,self.plotbl])
        elif ipol == 4:
            # The selected polarization is RR (not implemented because we only have amplitudes...)
            xx = self.msdata['data'][0,:,self.plotbl]
            yy = self.msdata['data'][1,:,self.plotbl]
            xy = self.msdata['data'][2,:,self.plotbl]
            yx = self.msdata['data'][3,:,self.plotbl]
            rr = (xx + yy + 1j*(xy - yx))/2
            spec = np.abs(rr)  
        elif ipol == 5:
            # The selected polarization is LL (not implemented because we only have amplitudes...)
            xx = self.msdata['data'][0,:,self.plotbl]
            yy = self.msdata['data'][1,:,self.plotbl]
            xy = self.msdata['data'][2,:,self.plotbl]
            yx = self.msdata['data'][3,:,self.plotbl]
            ll = (xx + yy - 1j*(xy - yx))/2
            spec = np.abs(ll)  

        i, j = blid[0] 
        if isauto[0]:
            self.ptitle = 'Auto-Correlation for Ant {}'.format(i+1)
        else:
            self.ptitle = 'Cross-Correlation for Baseline {}-{}'.format(i+1,j+1)
        spec1d = spec.flatten()
        spec1dg = np.sort(spec1d[~np.isnan(spec1d)])
        if len(spec1dg) == 0:
            self.statusBar.showMessage('This baseline is all NaN',2000)
            self.dspec_ax.set_title('This baseline is all NaN')
            self.speccanvas.draw()
        else:
            n95 = int(len(spec1dg)*0.95)
            ampmax = spec1dg[n95]
            ampmin = spec1dg[0]
            self.minentry.setText('{:10.2f}'.format(ampmin/10000.))
            self.maxentry.setText('{:10.2f}'.format(ampmax/10000.))
            self.plot_data()

    def plot_data(self):
        ''' Plot the data selected by get_data().
        '''
        # Determine which polarization state is selected by radiobuttons
        polnum = {'XX':0, 'YY':1, 'XY':2, 'YX':3, 'RR':4, 'LL':5}
        ipol = polnum[self.polGroup.checkedButton().text()]
        
        if ipol < 4:
            # The selected polarizatin is just one of the native polarizations (XX, YY, XY, or YX)
            spec = np.abs(self.msdata['data'][ipol,:,self.plotbl])
        elif ipol == 4:
            # The selected polarization is RR
            xx = self.msdata['data'][0,:,self.plotbl]
            yy = self.msdata['data'][1,:,self.plotbl]
            xy = self.msdata['data'][2,:,self.plotbl]
            yx = self.msdata['data'][3,:,self.plotbl]
            rr = (xx + yy + 1j*(xy - yx))/2
            spec = np.abs(rr)  
        elif ipol == 5:
            # The selected polarization is LL
            xx = self.msdata['data'][0,:,self.plotbl]
            yy = self.msdata['data'][1,:,self.plotbl]
            xy = self.msdata['data'][2,:,self.plotbl]
            yx = self.msdata['data'][3,:,self.plotbl]
            ll = (xx + yy - 1j*(xy - yx))/2
            spec = np.abs(ll)  

        try:
            minval = float(self.minentry.text())*10000.
        except:
            self.statusBar.showMessage('Could not interpret minimum sfu value',2000)
            self.minentry.setText('{:10.2f}'.format(np.nanmin(spec)/10000.))
            minval = float(self.minentry.text())*10000.
            
        try:
            maxval = float(self.maxentry.text())*10000.
        except:
            self.statusBar.showMessage('Could not interpret maximum sfu value',2000)
            self.maxentry.setText('{:10.2f}'.format(np.nanmax(spec)/10000.))
            maxval = float(self.maxentry.text())*10000.
            
        for i in range(len(self.dspec_ax.images)):
            # Clear any existing images without changing the axis limits
            self.dspec_ax.images[i].remove()
            
        fghz = self.msdata['freq']/1e9
        times = self.msdata['time']
        nf, nt = spec.shape
        if not self.ignore_gaps.isChecked():
            # Insert nan rows and columns in gaps
            df = np.median(fghz[1:] - fghz[:-1])
            fbreak, = np.where(fghz[1:] - fghz[:-1] > 2*df)
            for n,fb in enumerate(fbreak):
                loc = fb + n + 1
                fghz = np.concatenate((fghz[:loc],np.array([fghz[loc-1]+df]),fghz[loc:]))
                spec = np.concatenate((spec[:loc],np.zeros((1,nt))+np.nan,spec[loc:]),0)
            nf, nt = spec.shape
            dt = np.median(times[1:] - times[:-1])
            tbreak, = np.where(times[1:] - times[:-1] > 2*dt)
            for n,tb in enumerate(tbreak):
                loc = tb + n + 1
                times = np.concatenate((times[:loc],np.array([times[loc-1]+dt]),times[loc:]))
                spec = np.concatenate((spec[:,:loc],np.zeros((nf,1))+np.nan,spec[:,loc:]),1)
            nf, nt = spec.shape

        pd = Time(times/86400.,format='mjd').plot_date
        x0, x1, y0, y1 = self.dspec_ax.axis()   # Save current plot extent
        self.dspec_ax.clear()
        im = self.dspec_ax.pcolormesh(pd,fghz,np.clip(spec, minval, maxval))
        self.dspec_ax.xaxis_date()
        self.dspec_ax.xaxis.set_major_formatter(DateFormatter("%H:%M"))
        self.dspec_ax.set_ylabel('Frequency [GHz]')
        self.dspec_ax.set_xlabel('Time [UT]')
        if x0 == 0. and x1 == 1.:
            # Plot extent is 0,1 => cleared plot, so do not update plot extent
            pass
        else:
            # Restore saved plot extent
            self.dspec_ax.set_xlim(x0,x1)
            self.dspec_ax.set_ylim(y0,y1)
#        self.dspec_ax.imshow(np.clip(spec, minval, maxval), interpolation='nearest', aspect='auto', origin='lower')
        self.dspec_ax.set_title(self.ptitle)
        self.speccanvas.draw()
        
    def drawsel(self):
        ''' Uses timerange, spw, and stokes selection to draw a box on the plot.
        '''
        #if len(self.dspec_ax.get_images()) == 0:
        #    # No plot exists, so exit
        #    self.statusBar.showMessage("No plot exists yet. Please select a baseline to plot first.", 2000)
        #    return
        pd1, pd2 = trange2pd(self.trangeEdit.text())
        if pd1 is None:
            self.statusBar.showMessage(pd2, 2000)
            return
        f1, f2 = spw2frange(self.spwEdit.text(), self.msdata)
        if f1 is None:
            self.statusBar.showMessage(f2, 2000)
            return
        # Remove any existing lines before plotting a new one
        for line in self.dspec_ax.get_lines():
            line.remove()
        self.dspec_ax.plot([pd1, pd2, pd2, pd1, pd1],[f1, f1, f2, f2, f1],color='white')
        self.speccanvas.draw()
        
    def xferUL(self):
        ''' Uses the plot limits to define the timerange and spw selections.
        '''
        trange = pd2trange(self.dspec_ax.get_xlim())
        self.trangeEdit.setText(trange)
        f1, f2 = self.dspec_ax.get_ylim()
        idx1, = np.where(self.msdata['freq'] >= f1*1.e9)
        idx2, = np.where(self.msdata['freq'] <= f2*1.e9)
        spw = str(self.msdata['spwlist'][idx1[0]])+'~'+str(self.msdata['spwlist'][idx2[-1]])
        self.spwEdit.setText(spw)
    
    def xferDR(self):
        ''' Uses timerange, spw, and stokes selection to define the plot limits.
        '''
        pd = trange2pd(self.trangeEdit.text())
        if pd[0] is None:
            self.statusBar.showMessage(pd[1], 2000)
            return
        frange = spw2frange(self.spwEdit.text(), self.msdata)
        if frange[0] is None:
            self.statusBar.showMessage(frange[1], 2000)
            return
        self.dspec_ax.set_xlim(pd)
        self.dspec_ax.set_ylim(frange)            
        self.speccanvas.draw()
            
    def read_msdata(self, fillnan=None):
        ''' Read amplitudes for ALL data in the measurement set, returning
            a dictionary of the data, list of times, and list of frequencies.
            Fill flagged data with fillnan value if not None.
            
            Several changes to make this a LOT faster.  But nbl is hard-coded...
        '''
        ms = self.ms
        #if type(ms) != casatools.ms.ms:
        #    self.statusBar.showMessage('Not connected to a valid ms!',2000)
        #    return
            
        self.statusBar.addWidget(self.progressBar)
        self.progressBar.setFormat('Reading file...%p% done.')
        self.progressBar.setValue(0)
        self.progressBar.show()
        
#        ms.selectinit(datadescid=0, reset=True)
        spwinfo = ms.getspectralwindowinfo()
        nspw = len(spwinfo.keys())
#        spec = []
#        freq = []
        spwlist = []
#        for descid in range(len(spwinfo.keys())):
#            ms.selectinit(datadescid=0, reset=True)
#            ms.selectinit(datadescid=descid)
#            data = ms.getdata(['data', 'time', 'axis_info'], ifraxis=True)
#            spec_ = data['data']
#            freq_ = data['axis_info']['freq_axis']['chan_freq']
#            spwlist += [descid]*len(freq_)
#            if fillnan is not None:
#                flag_ = ms.getdata(['flag', 'time', 'axis_info'], ifraxis=True)['flag']
#                if type(fillnan) in [int, float]:
#                    spec_[flag_] = float(fillnan)
#                else:
#                    spec_[flag_] = 0.0
#            spec.append(spec_)
#            freq.append(freq_)
#            self.progressBar.setValue(100*(descid+1)/nspw)

        tb = table()
        tb.open(self.fname)
        spwtb = table()
        spwtb.open(self.fname+'/SPECTRAL_WINDOW')
        ptb = table()
        ptb.open(self.fname+'/POLARIZATION')
        
        mdata = ms.metadata()
        nspw = mdata.nspw()
        nbl = mdata.nbaselines() + mdata.nantennas()
        nscans = mdata.nscans()
        spw_nfrq = []     # List of number of frequencies in each spw
        for i in range(nspw):
            spw_nfrq.append(mdata.nchan(i))
        spw_nfrq = np.array(spw_nfrq)
        nf = np.sum(spw_nfrq)
        smry = mdata.summary()
        scan_ntimes = []  # List of number of times in each scan
        for iscan in range(nscans):
            scan_ntimes.append(len(smry['observationID=0']['arrayID=0']['scan='+str(iscan)]['fieldID=0'].keys()) - 6)
        scan_ntimes = np.array(scan_ntimes)
        nt = np.sum(scan_ntimes)
        times = tb.getcol('TIME')
        if times[nbl] - times[0] != 0:
            # This is frequency/scan sort order
            order = 'f'
        elif times[nbl*nspw] - times[0] !=0:
            # This is time sort order
            order = 't'
        npol = ptb.getcol('NUM_CORR',0,1)[0]
        ptb.close()
        freq = np.zeros(nf, float)
        times = np.zeros(nt, float)
        if order == 't':
            spec = np.zeros((npol, nf, nbl, nt), np.complex)
            for j in range(nt):
                fptr = 0
                # Loop over spw
                for i in range(nspw):
                    cfrq = spwtb.getcol('CHAN_FREQ',i,1)[:,0]     # Get channel frequencies for this spw (annoyingly comes out as shape (nf, 1)
                    if j == 0:
                        # Only need this the first time through
                        spwlist += [i]*len(cfrq)
                    if i == 0:
                        times[j] = tb.getcol('TIME',nbl*(i+nspw*j),1)  # Get the time
                    spec_ = tb.getcol('DATA',nbl*(i+nspw*j),nbl)  # Get complex data for this spw
                    flag = tb.getcol('FLAG',nbl*(i+nspw*j),nbl)   # Get flags for this spw
                    nfrq = len(cfrq)
                    # Apply flags
                    if type(fillnan) in [int, float]:
                        spec_[flag] = float(fillnan)
                    else:
                        spec_[flag] = 0.0
                    # Insert data for this spw into larger array
                    spec[:, fptr:fptr+nfrq, :, j] = spec_
                    freq[fptr:fptr+nfrq] = cfrq
                    fptr += nfrq
                self.progressBar.setValue(100*j/nt)
        else:
            specf = np.zeros((npol, nf, nt, nbl), np.complex)  # Array indexes are swapped
            iptr = 0
            for j in range(nscans):
                # Loop over scans
                for i in range(nspw):
                    #Loop over spectral windows
                    s = scan_ntimes[j]
                    f = spw_nfrq[i]
                    s1 = np.sum(scan_ntimes[:j])     # Start time index
                    s2 = np.sum(scan_ntimes[:j+1])   # End time index
                    f1 = np.sum(spw_nfrq[:i])         # Start freq index
                    f2 = np.sum(spw_nfrq[:i+1])       # End freq index
                    spec_ = tb.getcol('DATA',iptr,nbl*s)
                    flag  = tb.getcol('FLAG',iptr,nbl*s)
                    if j == 0:
                        cfrq = spwtb.getcol('CHAN_FREQ',i,1)[:,0]
                        freq[f1:f2] = cfrq
                        spwlist += [i]*len(cfrq)
                    times[s1:s2] = tb.getcol('TIME', iptr, nbl*s).reshape(s, nbl)[:,0]  # Get the times
                    iptr += nbl*s
                    # Apply flags
                    if type(fillnan) in [int, float]:
                        spec_[flag] = float(fillnan)
                    else:
                        spec_[flag] = 0.0
                    # Insert data for this spw into larger array
                    specf[:, f1:f2, s1:s2] = spec_.reshape(npol,f,s,nbl)
                    # Swap the array indexes back to the desired order
                    spec = np.swapaxes(specf,2,3)
                self.progressBar.setValue(100*j/nscans)
        tb.close()
        spwtb.close()

        self.statusBar.removeWidget(self.progressBar)
        self.dspec_ax.clear()  # Clear plot axis in preparation for a new dynamic spectrum
        self.msdata = {'data':spec, 'freq':freq, 'time':times, 'name':ms.name, 'spwlist':np.array(spwlist)}
                        
def get_bl2ord(nant=16):
    bl2ord = np.zeros((nant,nant),dtype='int')
    k = 0
    for i in range(nant):
        for j in range(i,nant):
            bl2ord[i,j] = k
            bl2ord[j,i] = k
            k += 1
    return bl2ord
    
def trange2pd(trange):
    ''' Convert a standard time range string to a pair of plot_date values.
        
        An error returns [None, msg]
    '''
    try:
        tstr1, tstr2 = trange.replace('/','-').split('~')
        return Time([tstr1[:10]+' '+tstr1[11:], tstr2[:10]+' '+tstr2[11:]]).plot_date
    except:
        return [None, 'Error interpreting timerange string']
        
def pd2trange(pd):
    ''' Convert a pair of plot_date values to a standard time trange string.
    
        An error returns None
    '''
    try:
        t1str = Time(pd[0],format='plot_date').iso.replace('-','/').replace(' ','/')
        t2str = Time(pd[1],format='plot_date').iso.replace('-','/').replace(' ','/')
        return t1str+'~'+t2str
    except:
        return None

def spw2frange(spw,msdata):
    ''' Convert a standard spectral window range to a pair of frequency values.
    '''
    try:
        spwrange = spw.split('~')
        if len(spwrange) == 1:
            spwrange *= 2
        spwlist = msdata['spwlist']
        fghz = msdata['freq']/1.e9
        idx1, = np.where(spwlist == int(spwrange[0]))
        idx2, = np.where(spwlist == int(spwrange[1]))
        return [fghz[idx1[0]], fghz[idx2[-1]]]
    except:
        return [None, 'Error interpreting spw string']

def pshift2pcenter(pshift,ms):
    ''' Convert phaseshift in heliocentric X,Y to RA, Dec using phase center in ms file

        pshift is a 2-element list of floats, e.g. [0.0, 0.0]

        output is the formatted RA Dec of the phase center, for use as phasecenter input in TCLEAN.
    '''
    from astropy.coordinates import SkyCoord
    from astropy import units as u
    from sunpy.coordinates import sun
    asec2rad = np.pi/(3600.*180.)
    
    mdata = ms.metadata()
    pcenter = mdata.phasecenter()
    # Phase center in radians
    ra0, dec0 = (pcenter['m0']['value'], pcenter['m1']['value'])
    trange = mdata.timerangeforobs(0)
    tstart = Time(trange['begin']['m0']['value'],format='mjd')  # Start time of file
    p0rad = sun.P(tstart).rad
    # Convert shifts in heliocentric to geocentric (i.e. rotate for solar P-angle)
    xgeo = (pshift[0]*np.cos(p0rad) - pshift[1]*np.sin(p0rad))*asec2rad
    ygeo = (pshift[0]*np.sin(p0rad) + pshift[1]*np.cos(p0rad))*asec2rad
    ra = ra0 - xgeo/np.cos(dec0)
    dec = dec0 + ygeo
    c = SkyCoord(ra=ra*u.rad, dec=dec*u.rad)
    rax = c.ra.hms
    decx = c.dec.dms
    phasecenter='J2000 {:02d}:{:02d}:{:07.4f} {:3d}.{:02d}.{:06.3f}'.format(int(rax.h),int(rax.m),rax.s,
                                                                          int(decx.d),int(abs(decx.m)),abs(decx.s))
    return phasecenter
        
if __name__ == '__main__':
    app = QApplication(sys.argv)
    ex = App()
    sys.exit(app.exec_())



import numpy as np
from matplotlib import transforms


def get_perp_vec(u1, u2, direction=1):
    """Return the unit vector perpendicular to the vector u2-u1."""

    x1, y1 = u1
    x2, y2 = u2
    vx, vy = x2 - x1, y2 - y1
    v = np.linalg.norm((vx, vy))
    wx, wy = -vy / v * direction, vx / v * direction
    return wx, wy


def get_av_vec(u1, u2):
    """Return the average unit vector between u1 and u2."""

    u1x, u1y = u1
    u2x, u2y = u2
    dx, dy = u1x + u2x, u1y + u2y
    dlen = np.linalg.norm((dx, dy))
    return dx / dlen, dy / dlen


class LineTicks:
    def __init__(self, line, idx, tick_length, tick_shift=3.0, direction=1, label=None, label_color='k',
                 fontsize='xx-small', **kwargs):
        self.line = line
        self.idx = idx
        self.tick_length = tick_length
        self.direction = direction
        self.label = label
        self.ticks = []
        self.tick_labels = []
        self.label_color = label_color
        self.fontsize = fontsize
        self.tick_shift = tick_shift
        self.tick_styles = kwargs
        # If no colour is specified for the ticks, set it to the line colour
        if not set(('c', 'color')).intersection(kwargs.keys()):
            self.tick_styles['color'] = self.line.get_color()

        self.ax = line.axes
        self.ax.callbacks.connect('xlim_changed', self.on_change_lims)
        self.ax.callbacks.connect('ylim_changed', self.on_change_lims)
        cid = self.ax.figure.canvas.mpl_connect('resize_event', self.on_resize)

        # get scale tick_length in display coordinates to data coordinates
        xy1, xy2 = self.ax.transData.inverted().transform([[0, 0], [self.tick_length, 0]])
        self.tick_length = xy2[0] - xy1[0]

        self.add_ticks(self.ax)

    def add_ticks(self, ax):
        ax.set_autoscale_on(False)  # Otherwise, infinite loop
        # Transform to  display coordinates
        # z =ax.transData.transform(np.array(self.line.get_data()).T)
        z = np.array(self.line.get_data()).T
        x, y = zip(*z)

        # Remove existing ticks
        for tick in self.ticks:
            ax.lines.remove(tick)
        # Remove references to the ticks so they can be garbage-collected
        self.ticks = []

        # Remove any existing tick labels
        for ticklabel in self.tick_labels:
            ax.texts.remove(ticklabel)
        self.tick_labels = []

        for j, i in enumerate(self.idx):
            if i == 0:
                # The first tick is perpendicular to the line between the
                # first two points
                tx, ty = get_perp_vec((x[0], y[0]), (x[1], y[1]),
                                      self.direction)
            elif i == len(x) - 1:
                # The last tick is perpendicular to the line between the
                # last two points
                tx, ty = get_perp_vec((x[-2], y[-2]), (x[-1], y[-1]),
                                      self.direction)
            else:
                # General tick marks bisect the incoming and outgoing line
                # segments
                u1 = get_perp_vec((x[i - 1], y[i - 1]), (x[i], y[i]),
                                  self.direction)
                u2 = get_perp_vec((x[i], y[i]), (x[i + 1], y[i + 1]),
                                  self.direction)
                tx, ty = get_av_vec(u1, u2)
            tx, ty = self.tick_length * tx, self.tick_length * ty
            this_tick, = ax.plot((x[i], x[i] + tx), (y[i], y[i] + ty), **self.tick_styles)
            self.ticks.append(this_tick)

            if self.label:
                this_ticklabel = ax.text(x[i] + tx * self.tick_shift, y[i] + ty * self.tick_shift, self.label[j],
                                         ha='center', va='center', clip_on=True, color=self.label_color,
                                         fontsize=self.fontsize)
                self.tick_labels.append(this_ticklabel)

    def on_change_lims(self, ax):
        self.add_ticks(ax)

    def on_resize(self, event):
        self.add_ticks(self.ax)



from dateutil import parser
from astropy.time import Time
from sunpy import map as smap
from scipy.io import readsav
import numpy as np

def idlsav2sunmap(idlsavfile):

    smaps_struc = readsav(idlsavfile)
    smaps_dict = {}
    for smap_key in smaps_struc.keys():
        smap_struc = smaps_struc[smap_key]
        smaps_dict[smap_key] = []
        smap_struc = np.squeeze(smap_struc)
        for tidx, mp in enumerate(smap_struc):
            header = {}
            hd_keys = mp.dtype.names
            for idx, k in enumerate(hd_keys):
                if k == 'DATA':
                    data = mp[k]
                elif k == 'DUR':
                    header['EXPTIME'] = mp[k]
                elif k == 'TIME':
                    header['DATE-OBS'] = Time(parser.parse(mp[k])).isot
                    header['DATE'] = header['DATE-OBS']
                elif k == 'XC':
                    header['CRVAL1'] = mp[k]
                elif k == 'DX':
                    header['CDELT1'] = mp[k]
                elif k == 'XUNITS':
                    header['CUNIT1'] = mp[k]
                elif k == 'L0':
                    header['HGLN_OBS'] = mp[k]
                elif k == 'YC':
                    header['CRVAL2'] = mp[k]
                elif k == 'DY':
                    header['CDELT2'] = mp[k]
                elif k == 'RSUN_OBS':
                    header['CDELT2'] = mp[k]
                elif k == 'YUNITS':
                    header['CUNIT2'] = mp[k]
                elif k == 'B0':
                    header['HGLT_OBS'] = mp[k]
                elif k == 'FREQ':
                    header['WAVELNTH'] = mp[k]
                elif k == 'FREQUNIT':
                    header['WAVEUNIT'] = mp[k]
                else:
                    header[k] = mp[k]
            header['CTYPE1'] = 'HPLN-TAN'
            header['CTYPE2'] = 'HPLT-TAN'
            header['CRPIX2'], header['CRPIX1'] = (np.array(data.shape) + 1.0) / 2 + 0.5
            header['TELESCOP'] = 'EOVSA'
            header['RSUN_REF'] = 695508000.0
            header['P_ANGLE'] = 0.0
            header['NAXIS'] = 2
            header['NAXIS2'], header['NAXIS1'] = data.shape
            sunmap = smap.Map(data, header)
            smaps_dict[smap_key].append(sunmap)
    return smaps_dict


'''
   Module for converting CASA image analysis outputs to data recognizable by DataBrowser   
'''
# History:
#   2016-Nov-2 BC
#       New module

import numpy as np

def pimfitres2dict(pimfitres):
    ''' parse pimfit results to a dictionary '''
    # check whether the input is a single dictionary (from imfit) or a list (from pimfit)
    if isinstance(imfitres,dict):
        imfitres=[imfitres]
    if (not isinstance(imfitres,list)):
        print 'input "imfitres" is not a list. Abort...'
    out = []
    ntim = len(imfitres)
    for i, res in enumerate(imfitres):
        if 'converged' in res:
            nfreq=len(res['converged'])


         




import sunpy.map as smap
from dateutil import parser
from astropy.time import Time
import numpy as np


def map2smap(sswmap):
    def mp2smap(sswmp):
        header = {}
        hd_keys = sswmp.dtype.names
        for idx, k in enumerate(hd_keys):
            if k == 'DATA':
                data = sswmp[idx]
            elif k == 'DUR':
                header['EXPTIME'] = sswmp[idx]
            elif k == 'TIME':
                header['DATE-OBS'] = Time(parser.parse(sswmp[idx])).isot
                header['DATE'] = header['DATE-OBS']
            elif k == 'XC':
                header['CRVAL1'] = sswmp[idx]
            elif k == 'DX':
                header['CDELT1'] = sswmp[idx]
            elif k == 'XUNITS':
                cunit1 = sswmp[idx]
                if isinstance(cunit1, bytes):
                    cunit1 = cunit1.decode('utf-8')
                if cunit1 == 'arcsecs':
                    header['CUNIT1'] = 'arcsec'
                else:
                    header['CUNIT1'] = cunit1
            elif k == 'L0':
                header['HGLN_OBS'] = sswmp[idx]
            elif k == 'YC':
                header['CRVAL2'] = sswmp[idx]
            elif k == 'DY':
                header['CDELT2'] = sswmp[idx]
            elif k == 'RSUN_OBS':
                header['CDELT2'] = sswmp[idx]
            elif k == 'YUNITS':
                cunit1 = sswmp[idx]
                if isinstance(cunit1, bytes):
                    cunit1 = cunit1.decode('utf-8')
                if cunit1 == 'arcsecs':
                    header['CUNIT2'] = 'arcsec'
                else:
                    header['CUNIT2'] = cunit1
            elif k == 'B0':
                header['HGLT_OBS'] = sswmp[idx]
            elif k == 'ID':
                telescop = sswmp[idx]
                if isinstance(telescop, bytes):
                    telescop = telescop.decode('utf-8')
                if 'SDO AIA' in telescop:
                    header['TELESCOP'] = 'SDO AIA'
                elif 'SDO HMI' in telescop:
                    header['TELESCOP'] = 'SDO HMI'
                elif 'EOVSA' in telescop:
                    header['TELESCOP'] = 'EOVSA'
                elif 'VLA' in telescop:
                    header['TELESCOP'] = 'VLA'
                else:
                    header['TELESCOP'] = telescop
            else:
                header[k] = sswmp[idx]
        header['CTYPE1'] = 'HPLN-TAN'
        header['CTYPE2'] = 'HPLT-TAN'
        header['CRPIX2'], header['CRPIX1'] = (np.array(data.shape) + 1.0) / 2
        # header['TELESCOP'] = 'EOVSA'
        header['RSUN_REF'] = 695508000.0
        header['P_ANGLE'] = 0.0
        return smap.Map(data, header)

    if isinstance(sswmap, np.recarray):
        res = []
        for mp in sswmap:
            res.append(mp2smap(mp))
        return res
    elif isinstance(sswmap, np.record):
        return mp2smap(sswmap)



##################### generated by xml-casa (v2) from pimfit.xml ####################
##################### d7d637360366fa9a556c7f929f11bf8d ##############################
from __future__ import absolute_import
import numpy
from casatools.typecheck import CasaValidator as _val_ctor
_pc = _val_ctor( )
from casatools.coercetype import coerce as _coerce
from .private.task_pimfit import pimfit as _pimfit_t
from casatasks.private.task_logging import start_log as _start_log
from casatasks.private.task_logging import end_log as _end_log

class _pimfit:
    """
    pimfit ---- Fit one or more elliptical Gaussian components on an image region(s)

    --------- parameter descriptions ---------------------------------------------

    imagefiles   A list of the input images
    ncpu         Number of cpu cores to use
    doreg        True if use vla_prep to register the image
    ephemfile    emphemeris file generated from vla_prep.read_horizons()
    timestamps   A list of timestamps of the input images
    msinfofile   time-dependent phase center information generated from vla_prep.read_msinfo()
    box          Rectangular region(s) to select in direction plane. See "help par.box" for details. Default is to use the entire direction plane.
    region       Region selection. See "help par.region" for details. Default is to use the full image.
    chans        Channels to use. See "help par.chans" for details. Default is to use all channels.
    stokes       Stokes planes to use. See "help par.stokes" for details. Default is to use first Stokes plane.
    mask         Mask to use. See help par.mask. Default is none.
    includepix   Range of pixel values to include for fitting.
    excludepix   Range of pixel values to exclude for fitting.
    residual     Name of output residual image.
    model        Name of output model image.
    estimates    Name of file containing initial estimates of component parameters.
    logfile      Name of file to write fit results.
    append       If logfile exists, append to it if True or overwrite it if False
    newestimates File to write fit results which can be used as initial estimates for next run.
    complist     Name of output component list table.
    overwrite    Overwrite component list table if it exists?
    dooff        Also fit a zero level offset? Default is False
    offset       Initial estimate of zero-level offset. Only used if doff is True. Default is 0.0
    fixoffset    Keep the zero level offset fixed during fit? Default is False
    stretch      Stretch the mask if necessary and possible? See help par.stretch
    rms          RMS to use in calculation of uncertainties. Numeric or valid quantity (record or string). If numeric, it is given units of the input image. If quantity, units must conform to image units. If not positive, the rms of the residual image, in the region of the fit, is used.
    noisefwhm    Noise correlation beam FWHM. If numeric value, interpreted as pixel widths. If quantity (dictionary, string), it must have angular units.
    summary      File name to which to write table of fit parameters.
    [1;42mRETURNS[1;m         void

    --------- examples -----------------------------------------------------------

    
    PARAMETER SUMMARY
    imagename        Name of the input image
    box              Rectangular region(s) to select in direction plane. See "help par.box"
    for details. Default is to use the entire direction plane.
    eg "100, 120, 200, 220, 300, 300, 400, 400" to use two boxes.
    region           Region selection. See "help par.region" for details. Default is to use
    the full image.
    chans            Channels to use. See "help par.chans" for details. Default is to use all
    channels.
    stokes           Stokes planes to use. See "help par.stokes" for details. Default is to
    use first Stokes plane.
    mask             Mask to use. See help par.mask. Default is none.
    includepix       Range of pixel values to include for fitting. Array of two numeric
    values assumed to have same units as image pixel values. Only one
    of includepix or excludepix can be specified.
    excludepix       Range of pixel values to exclude for fitting. Array of two numeric
    values assumed to have same units as image pixel values. Only one
    of includepix or excludepix can be specified.
    residual         Name of the residual image to write.
    model            Name of the model image to write.
    estimates        Name of file containing initial estimates of component parameters
    (see below for formatting details).
    logfile          Name of file to write fit results.
    append           If logfile exists, append to it (True) or overwrite it (False).
    newestimates     File to write fit results which can be used as initial estimates
    for next run.
    complist         Name of output component list table.
    overwrite        Overwrite component list table if it exists?
    dooff            Simultaneously fit a zero-level offset?
    offset           Initial estimate for the zero-level offset. Only used if dooff is True.
    fixoffset        Hold zero-level offset constant during fit? Only used if dooff is True.
    stretch          Stretch the input mask if necessary and possible. Only used if a mask is specified.
    See help par.stretch.
    rms              RMS to use in calculation of various uncertainties, assumed to have units of the input
    image. If not positve, the rms of the residual image is used.
    noisefwhm        Noise correlation beam FWHM. If numeric value, interpreted as pixel widths. If
    quantity (dictionary, string), it must have angular units.
    summary          File name to which to write table of fit parameters.
    
    OVERVIEW
    This application is used to fit one or more two dimensional gaussians to sources in an image as
    well as an optional zero-level offset. Fitting is limited to a single polarization
    but can be performed over several contiguous spectral channels.
    If the image has a clean beam, the report and returned dictionary will contain both the convolved
    and the deconvolved fit results.
    
    When dooff is False, the method returns a dictionary with three keys, 'converged', 'results',
    and 'deconvolved'. The value of 'converged' is a boolean array which indicates if the fit
    converged on a channel by channel basis. The value of 'results' is a dictionary representing
    a component list reflecting the fit results. In the case of an image containing beam information,
    the sizes and position angles in the 'results' dictionary are those of the source(s) convolved
    with the restoring beam, while the same parameters in the 'deconvolved' dictionary represent the
    source sizes deconvolved from the beam. In the case where the image does not contain a beam,
    'deconvolved' will be absent. Both the 'results' and 'deconvolved' dictionaries can
    be read into a component list tool (default tool is named cl) using the fromrecord() method
    for easier inspection using tool methods, eg
    
    cl.fromrecord(res['results'])
    
    although this currently only works if the flux density units are conformant with Jy.
    
    There are also values in each component subdictionary not used by cl.fromrecord() but meant to
    supply additional information. There is a 'peak' subdictionary for each component that provides the
    peak intensity of the component. It is present for both 'results' and 'deconvolved' components.
    There is also a 'sum' subdictionary for each component indicated the simple sum of pixel values in
    the the original image enclosed by the fitted ellipse. There is a 'channel' entry in the 'spectrum'
    subdictionary which provides the zero-based channel number in the input image for which the solution
    applies. In addtion, if the image has a beam(s), then there will be a 'beam' subdictionary associated
    with each component in both the 'results' and 'deconvolved' dictionaries. This subdictionary will
    have three keys: 'beamarcsec' will be a subdictionary giving the beam dimensions in arcsec,
    'beampixels' will have the value of the beam area expressed in pixels, and 'beamster' will have the
    value of the beam area epressed in steradians. Also, if the image has a beam(s), in the component level
    dictionaries will be an 'ispoint' entry with an associated boolean value describing if the component
    is consistent with a point source.
    
    If dooff is True, in addtion to the specified number of
    gaussians, a zero-level offset will also be fit. The initial estimate for this
    offset is specified using the offset parameter. Units are assumed to be the
    same as the image brightness units. The zero level offset can be held constant during
    the fit by specifying fixoffset=True. In the case of dooff=True, the returned
    dictionary contains two additional keys, 'zerooff' and 'zeroofferr', which are both
    dictionaries containing 'unit' and 'value' keys. The values associated with the 'value'
    keys are arrays containing the the fitted zero level offset value and its error, respectively,
    for each channel. In cases where the fit did not converge, these values are set to NaN.
    The value associated with 'unit' is just the i`mage brightness unit.
    
    The region can either be specified by a box(es) or a region.
    Ranges of pixel values can be included or excluded from the fit. If specified using
    the box parameter, multiple boxes can be given using the format
    box="blcx1, blcy1, trcx1, trcy1, blcx2, blcy2, trcx2, trcy2, ... , blcxN, blcyN, trcxN, trcyN"
    where N is the number of boxes. In this case, the union of the specified boxes will be used.
    
    If specified, the residual and/or model images for successful fits will be written.
    
    If an estimates file is not specified, an attempt is made to estimate
    initial parameters and fit a single Gaussian. If a multiple Gaussian fit
    is desired, the user must specify initial estimates via a text file
    (see below for details).
    
    The user has the option of writing the result of the fit to a log file,
    and has the option of either appending to or overwriting an existing file.
    
    The user has the option of writing the (convolved) parameters of a successful
    fit to a file which can be fed back to fitcomponents() as the estimates file for a
    subsequent run.
    
    The user has the option of writing the fit results in tabular format to a file whose
    name is specified using the summary parameter.
    
    If specified and positive, the value of rms is used to calculate the parameter uncertainties,
    otherwise, the rms in the selected region in the relevant channel is used for these calculations.
    
    The noisefwhm parameter represents the noise-correlation beam FWHM. If specified as a quantity,
    it should have angular units. If specified as a numerical value, it is set equal to that number
    of pixels. If specified and greater than or equal to the pixel size, it is used to calculate
    parameter uncertainties using the correlated noise equations (see below). If it is specified but
    less than a pixel width, the the uncorrelated noise equations (see below) are used to
    compute the parameter uncertainties. If it is not specified and the image has a restoring beam(s),
    the the correlated noise equations are used to compute parameter uncertainties using the
    geometric mean of the relevant beam major and minor axes as the noise-correlation beam FWHM. If
    noisefwhm is not specified and the image does not have a restoring beam, then the uncorrelated
    noise equations are used to compute the parameter uncertainties.
    
    SUPPORTED UNITS
    
    Currently only images with brightness units conformant with Jy/beam, Jy.km/s/beam, and K are fully
    supported for fitting. If your image has some other base brightness unit, that unit will be assumed
    to be equivalent to Jy/pixel and results will be calculated accordingly. In particular,
    the flux density (reported as Integrated Flux in the logger and associated with the "flux" key
    in the returned component subdictionary(ies)) for such a case represents the sum of pixel values.
    
    Note also that converting the returned results subdictionary to a component list via cl.fromrecord() currently
    only works properly if the flux density units in the results dictionary are conformant with Jy.
    If you need to be able to run cl.fromrecord() on the resulting dictionary you can first modify the
    flux density units by hand to be (some prefix)Jy and then run cl.fromrecord() on that dictionary,
    bearing in mind your unit conversion.
    
    If the input image has units of K, the flux density of components will be reported in units
    of [prefix]K*rad*rad, where prefix is an SI prefix used so that the numerical value is between
    1 and 1000. To convert to units of K*beam, determine the area of the appropriate beam,
    which is given by pi/(4*ln(2))*bmaj*bmin, where bmaj and bmin are the major and minor axes
    of the beam, and convert to steradians (=rad*rad). This value is included in the beam portion
    of the component subdictionary (key 'beamster'). Then divide the numerical value of the
    logged flux density by the beam area in steradians. So, for example
    
    begin{verbatim}
    # run on an image with K brightness units
    res = imfit(...)
    # get the I flux density in K*beam of component 0
    comp = res['results']['component0']
    flux_density_kbeam = comp['flux']['value'][0]/comp['beam']['beamster']
    end{verbatim}
    
    FITTING OVER MULTIPLE CHANNELS
    
    For fitting over multiple channels, the result of the previous successful fit is used as
    the estimate for the next channel. The number of gaussians fit cannot be varied on a channel
    by channel basis. Thus the variation of source structure should be reasonably smooth in
    frequency to produce reliable fit results.
    
    MASK SPECIFICATION
    
    Mask specification can be done using an LEL expression. For example
    
    mask = '"myimage">5' will use only pixels with values greater than 5.
    
    INCLUDING AND EXCLUDING PIXELS
    
    Pixels can be included or excluded from the fit based on their values
    using these parameters. Note that specifying both is not permitted and
    will cause an error. If specified, both take an array of two numeric
    values.
    
    ESTIMATES
    
    Initial estimates of fit parameters may be specified via an estimates
    text file. Each line of this file should contain a set of parameters for
    a single gaussian. Optionally, some of these parameters can be fixed during
    the fit. The format of each line is
    
    peak intensity, peak x-pixel value, peak y-pixel value, major axis, minor axis, position angle, fixed
    
    The fixed parameter is optional. The peak intensity is assumed to be in the
    same units as the image pixel values (eg Jy/beam). The peak coordinates are specified
    in pixel coordinates. The major and minor axes and the position angle are the convolved
    parameters if the image has been convolved with a clean beam and are specified as quantities.
    The fixed parameter is optional and is a string. It may contain any combination of the
    following characters 'f' (peak intensity), 'x' (peak x position), 'y' (peak y position),
    'a' (major axis), 'b' (minor axis), 'p' (position angle).
    
    In addition, lines in the file starting with a # are considered comments.
    
    An example of such a file is:
    
    begin{verbatim}
    # peak intensity must be in map units
    120, 150, 110, 23.5arcsec, 18.9arcsec, 120deg
    90, 60, 200, 46arcsec, 23arcsec, 140deg, fxp
    end{verbatim}
    
    
    This is a file which specifies that two gaussians are to be simultaneously fit,
    and for the second gaussian the specified peak intensity, x position, and position angle
    are to be held fixed during the fit.
    
    ERROR ESTIMATES
    
    Error estimates are based on the work of Condon 1997, PASP, 109, 166. Key assumptions made are:
    * The given model (elliptical Gaussian, or elliptical Gaussian plus constant offset) is an
    adequate representation of the data
    * An accurate estimate of the pixel noise is provided or can be derived (see above). For the
    case of correlated noise (e.g., a CLEAN map), the fit region should contain many "beams" or
    an independent value of rms should be provided.
    * The signal-to-noise ratio (SNR) or the Gaussian component is large. This is necessary because
    a Taylor series is used to linearize the problem. Condon (1997) states that the fractional
    bias in the fitted amplitude due to this assumption is of order 1/(S*S), where S is the overall
    SNR of the Gaussian with respect to the given data set (defined more precisely below). For a 5
    sigma "detection" of the Gaussian, this is a 4% effect.
    * All (or practically all) of the flux in the component being fit falls within the selected region.
    If a constant offset term is simultaneously fit and not fixed, the region of interest should be
    even larger. The derivations of the expressions summarized in this note assume an effectively
    infinite region.
    
    Two sets of equations are used to calculate the parameter uncertainties, based on if
    the noise is correlated or uncorrelated. The rules governing which set of equations are
    used have been described above in the description of the noisefwhm parameter.
    
    In the case of uncorrelated noise, the equations used are
    
    f(A) = f(I) = f(M) = f(m) = k*s(x)/M = k*s(y)/m = (s(p)/sqrt(2))*((M*M - m*m)/(M*m))
    = sqrt(2)/S
    
    where s(z) is the uncertainty associated with parameter z, f(z) = s(z)/abs(z) is the
    fractional uncertainty associated with parameter z, A is the peak intensity, I is the flux
    density, M  and m are the FWHM major and minor axes, p is the position angle of the
    component, and k = sqrt(8*ln(2)). s(x) and s(y) are the direction
    uncertainties of the component measured along the major and minor axes; the resulting
    uncertainties measured along the principle axes of the image direction coordinate are
    calculated by propagation of errors using the 2D rotation matrix which enacts the rotation through
    the position angle plus 90 degrees. S is the overall signal to noise ratio of the component,
    which, for the uncorrelated noise case is given by
    
    S = (A/(k*h*r))*sqrt(pi*M*m)
    
    where h is the pixel width of the direction coordinate and r is the rms noise (see the
    discussion above for the rules governing how the value of r is determined).
    
    For the correlated noise case, the same equations are used to determine the uncertainties
    as in the uncorrelated noise case, except for the uncertainty in I (see below). However,
    S is given by
    
    S = (A/(2*r*N)) * sqrt(M*m) * (1 + ((N*N/(M*M)))**(a/2)) * (1 + ((N*N/(m*m)))**(b/2))
    
    where N is the noise-correlation beam FWHM (see discussion of the noisefwhm parameter for
    rules governing how this value is determined). "**" indicates exponentiation and a and b
    depend on which uncertainty is being calculated. For sigma(A), a = b = 3/2. For M and x,
    a = 5/2 and b = 1/2. For m, y, and p, a = 1/2 and b = 5/2. f(I) is calculated in the
    correlated noise case according to
    
    f(I) = sqrt( f(A)*f(A) + (N*N/(M*m))*(f(M*f(M) + f(m)*f(m))) )
    
    Note well the following caveats:
    * Fixing Gaussian component parameters will tend to cause the parameter uncertainties reported for free
    parameters to be overestimated.
    * Fitting a zero level offset that is not fixed will tend to cause the reported parameter
    uncertainties to be slightly underestimated.
    * The parameter uncertainties will be inaccurate at low SNR (a ~10% for SNR = 3).
    * If the fitted region is not considerably larger than the largest component that is fit,
    parameter uncertainties may be mis-estimated.
    * An accurate rms noise measurement, r, for the region in question must be supplied.
    Alternatively, a sufficiently large signal-free region must be present in the selected region
    (at least about 25 noise beams in area) to auto-derive such an estimate.
    * If the image noise is not statistically independent from pixel to pixel, a reasonably accurate noise
    correlation scale, N, must be provided. If the noise correlation function is not approximately Gaussian,
    the correlation length can be estimated using
    
    N = sqrt(2*ln(2)/pi)* double-integral(dx dy C(x,y))/sqrt(double-integral(dx dy C(x, y) * C(x,y)))
    
    where C(x,y) is the associated noise-smoothing function
    * If fitted model components have significan spatial overlap, the parameter uncertainties are likely to
    be mis-estimated (i.e., correlations between the parameters of separate components are not accounted
    for).
    * If the image being analyzed is an interferometric image with poor uv sampling, the parameter
    uncertainties may be significantly underestimated.
    
    The deconvolved size and position angle errors are computed by taking the maximum of the absolute values of the
    differences of the best fit deconvolved value of the given parameter and the deconvolved size of the eight
    possible combinations of (FWHM major axis +/- major axis error), (FWHM minor axis +/- minor axis error),
    and (position andle +/- position angle error). If the source cannot be deconvolved from the beam (if the best
    fit convolved source size cannot be deconvolved from the beam), upper limits on the deconvolved source size
    are sometimes reported. These limits simply come from the maximum major and minor axes of the deconvolved
    gaussians taken from trying all eight of the aforementioned combinations. In the case none of these combinations
    produces a deconvolved size, no upper limit is reported.
    
    EXAMPLE:
    
    Here is how one might fit two gaussians to multiple channels of a cube using the fit
    from the previous channel as the initial estimate for the next. It also illustrates
    how one can specify a region in the associated continuum image as the region to use
    as the fit for the channel.
    
    
    begin{verbatim}
    default imfit
    imagename = "co_cube.im"
    # specify region using region from continuum
    region = "continuum.im:source.rgn"
    chans = "2~20"
    # only use pixels with positive values in the fit
    excludepix = [-1e10,0]
    # estimates file contains initial parameters for two Gaussians in channel 2
    estimates = "initial_estimates.txt"
    logfile = "co_fit.log"
    # append results to the log file for all the channels
    append = "True"
    imfit()


    """

    _info_group_ = """analysis"""
    _info_desc_ = """Fit one or more elliptical Gaussian components on an image region(s)"""

    def __call__( self, imagefiles=[  ], ncpu=int(8), doreg=False, ephemfile='', timestamps=[  ], msinfofile='', box='', region='', chans='', stokes='', mask='', includepix=[  ], excludepix=[  ], residual='', model='', estimates='', logfile='', append=True, newestimates='', complist='', overwrite=False, dooff=False, offset=float(0.0), fixoffset=False, stretch=False, rms=int(0), noisefwhm='', summary='' ):
        schema = {'imagefiles': {'type': 'cReqPathVec', 'coerce': [_coerce.to_list,_coerce.expand_pathvec]}, 'ncpu': {'type': 'cInt'}, 'doreg': {'type': 'cBool'}, 'ephemfile': {'type': 'cStr', 'coerce': _coerce.to_str}, 'timestamps': {'type': 'cStrVec', 'coerce': [_coerce.to_list,_coerce.to_strvec]}, 'msinfofile': {'type': 'cStr', 'coerce': _coerce.to_str}, 'box': {'type': 'cStr', 'coerce': _coerce.to_str}, 'region': {'type': 'cVariant', 'coerce': [_coerce.to_variant]}, 'chans': {'type': 'cVariant', 'coerce': [_coerce.to_variant]}, 'stokes': {'type': 'cStr', 'coerce': _coerce.to_str}, 'mask': {'type': 'cStr', 'coerce': _coerce.to_str}, 'includepix': {'type': 'cIntVec', 'coerce': [_coerce.to_list,_coerce.to_intvec]}, 'excludepix': {'type': 'cIntVec', 'coerce': [_coerce.to_list,_coerce.to_intvec]}, 'residual': {'type': 'cStr', 'coerce': _coerce.to_str}, 'model': {'type': 'cStr', 'coerce': _coerce.to_str}, 'estimates': {'type': 'cStr', 'coerce': _coerce.to_str}, 'logfile': {'type': 'cStr', 'coerce': _coerce.to_str}, 'append': {'type': 'cBool'}, 'newestimates': {'type': 'cStr', 'coerce': _coerce.to_str}, 'complist': {'type': 'cStr', 'coerce': _coerce.to_str}, 'overwrite': {'type': 'cBool'}, 'dooff': {'type': 'cBool'}, 'offset': {'type': 'cFloat', 'coerce': _coerce.to_float}, 'fixoffset': {'type': 'cBool'}, 'stretch': {'type': 'cBool'}, 'rms': {'anyof': [{'type': 'cInt'}, {'type': 'cFloat', 'coerce': _coerce.to_float}, {'type': 'cDict'}, {'type': 'cStr', 'coerce': _coerce.to_str}]}, 'noisefwhm': {'anyof': [{'type': 'cInt'}, {'type': 'cFloat', 'coerce': _coerce.to_float}, {'type': 'cDict'}, {'type': 'cStr', 'coerce': _coerce.to_str}]}, 'summary': {'type': 'cStr', 'coerce': _coerce.to_str}}
        doc = {'imagefiles': imagefiles, 'ncpu': ncpu, 'doreg': doreg, 'ephemfile': ephemfile, 'timestamps': timestamps, 'msinfofile': msinfofile, 'box': box, 'region': region, 'chans': chans, 'stokes': stokes, 'mask': mask, 'includepix': includepix, 'excludepix': excludepix, 'residual': residual, 'model': model, 'estimates': estimates, 'logfile': logfile, 'append': append, 'newestimates': newestimates, 'complist': complist, 'overwrite': overwrite, 'dooff': dooff, 'offset': offset, 'fixoffset': fixoffset, 'stretch': stretch, 'rms': rms, 'noisefwhm': noisefwhm, 'summary': summary}
        assert _pc.validate(doc,schema), str(_pc.errors)
        _logging_state_ = _start_log( 'pimfit', [ 'imagefiles=' + repr(_pc.document['imagefiles']), 'ncpu=' + repr(_pc.document['ncpu']), 'doreg=' + repr(_pc.document['doreg']), 'ephemfile=' + repr(_pc.document['ephemfile']), 'timestamps=' + repr(_pc.document['timestamps']), 'msinfofile=' + repr(_pc.document['msinfofile']), 'box=' + repr(_pc.document['box']), 'region=' + repr(_pc.document['region']), 'chans=' + repr(_pc.document['chans']), 'stokes=' + repr(_pc.document['stokes']), 'mask=' + repr(_pc.document['mask']), 'includepix=' + repr(_pc.document['includepix']), 'excludepix=' + repr(_pc.document['excludepix']), 'residual=' + repr(_pc.document['residual']), 'model=' + repr(_pc.document['model']), 'estimates=' + repr(_pc.document['estimates']), 'logfile=' + repr(_pc.document['logfile']), 'append=' + repr(_pc.document['append']), 'newestimates=' + repr(_pc.document['newestimates']), 'complist=' + repr(_pc.document['complist']), 'overwrite=' + repr(_pc.document['overwrite']), 'dooff=' + repr(_pc.document['dooff']), 'offset=' + repr(_pc.document['offset']), 'fixoffset=' + repr(_pc.document['fixoffset']), 'stretch=' + repr(_pc.document['stretch']), 'rms=' + repr(_pc.document['rms']), 'noisefwhm=' + repr(_pc.document['noisefwhm']), 'summary=' + repr(_pc.document['summary']) ] )
        return _end_log( _logging_state_, 'pimfit', _pimfit_t( _pc.document['imagefiles'], _pc.document['ncpu'], _pc.document['doreg'], _pc.document['ephemfile'], _pc.document['timestamps'], _pc.document['msinfofile'], _pc.document['box'], _pc.document['region'], _pc.document['chans'], _pc.document['stokes'], _pc.document['mask'], _pc.document['includepix'], _pc.document['excludepix'], _pc.document['residual'], _pc.document['model'], _pc.document['estimates'], _pc.document['logfile'], _pc.document['append'], _pc.document['newestimates'], _pc.document['complist'], _pc.document['overwrite'], _pc.document['dooff'], _pc.document['offset'], _pc.document['fixoffset'], _pc.document['stretch'], _pc.document['rms'], _pc.document['noisefwhm'], _pc.document['summary'] ) )

pimfit = _pimfit( )

